/* This connection library is the SNIPE_Lite code from the UTK/ORNL SNIPE
   project. G.Fagg 1998- email fagg@cs.utk.edu */

/* Modifications by Kevin London london@cs.utk.edu
 * Specifically, picking random port, and dialing up/down send and
 * Recv buffers
 *
 * Added windows32 support -Kevin London london@cs.utk.edu
 */

#include <sys/types.h>
#ifndef WIN32
#include <sys/param.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <strings.h>
#include <sys/time.h>
#else
#include <Winsock2.h>
#include <windows.h>
#include <time.h>
#include <sys/timeb.h>
#endif
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <malloc.h>

#define RETRIES 2

#define socklen_t int

#undef DEBUG
#undef VERBOSE

/* now that we have included everything under the Sun... */
/* and I won't mention Sun TLI problems here */


void nodelay();	/* hack routine to improve performance.. maybe */
void setnonblocking(); /* sets the given socket non-blocking */
void setblocking(); /* sets the given socket blocking again */
int setsendrecvbufs();  /* Sets send/recv buffer sizes */
int setsendbuf(); /* Sets send buffer size */
int setrecvbuf(); /* Sets recv buffer size */


int setportconn (s, port, range)
    /* returns final port value or error */
    int *s;					/* socket server uses.. this is created here */
int port; 				/* initial port we make service on. */
int range;				/* range of ports */
{
    struct sockaddr_in  sa;
    int ts;
    int rc,i;
#ifndef WIN32
    int p;
    size_t oslen;
#else
    unsigned short p;
    int oslen;
    WORD wVersionRequested;
    WSADATA wsaData;

    wVersionRequested = MAKEWORD( 2, 0 );
    WSAStartup ( wVersionRequested, &wsaData );
#endif

    p = port;
    ts = socket (AF_INET, SOCK_STREAM, 0);   /* yuck */
#ifdef VERBOSE
    printf("Socket [%d]\n", ts);
#endif /* VERBOSE */
    sa.sin_family       = AF_INET;
    sa.sin_addr.s_addr  = INADDR_ANY;
    oslen = sizeof(sa);
    sa.sin_port         = htons (p); /* convert just to be safe */
#ifdef VERBOSE
	printf ("Port [%d] NB [0x%x]\n", p, sa.sin_port);
#endif /* VERBOSE */
    rc = ( bind(ts, (struct sockaddr *)&sa, sizeof (sa)));
    if ( rc ){
#ifdef VERBOSE
       printf("Port [%d] in use.\n", p );
#endif
       for ( i=1;i<6;i++) {
		 sa.sin_port = htons((unsigned short)(port+i));
         rc = ( bind(ts, (struct sockaddr *)&sa, sizeof (sa)));
	 if ( !rc )
		break;
       }
       if ( rc ) {
#ifdef VERBOSE
	printf("Using a RANDOM port!\n");
#endif
       	sa.sin_port  = htons( 0 );
       	rc = ( bind(ts, (struct sockaddr *)&sa, sizeof (sa)));
       }
    }
#ifdef VERBOSE
    else
       printf("Bind on port [%d] succesful.\n", p );
#endif

    if ( !rc ) {
        getsockname( ts, (struct sockaddr *)&sa, &oslen); 
	p = ntohs(sa.sin_port);  /* Need to get the real number -KSL */
#ifdef VERBOSE
	printf ("Bind on socket [%d] port [%d] = [%d]\n", ts, p, rc);
	printf("Attempting listen after bind on socket [%d]\n", ts);
#endif /* VERBOSE */
	rc = listen (ts, 25);	/* handle 5 possible connection at once */
#ifdef VERBOSE
	printf("Listen on socket [%d] = [%d]\n", rc, ts);
#endif /* VERBOSE */
	if (rc) perror ("Listen()");
	if (!rc) { 
	     *s = ts;	/* return socket value as well */
	      return (p);
        }
    }

    /* ok.. we didnot find one */
#ifndef WIN32
    close (ts);
#else
    closesocket(ts);
#endif
    *s = 0;	/* return no socket */
    return (rc);	/* which is negative and the last bind return code */
}

int allowconn (s, fe, spid)
    int s;					/* socket we listen on */
int fe;					/* Fork Enabled? 0=no 1=yes */
/* If enabled then forks new process on accept */
int *spid;				/* new process id handling the returned socked */
/* used by multitasking servers */
{
    int s2;	/* new socket */
    /* struct sockaddr *new_s_addr_ptr;  */
    struct sockaddr new_s_addr; 
#ifndef WIN32
    size_t new_s_addr_len;
#else
    int new_s_addr_len;
#endif


    new_s_addr_len = sizeof (new_s_addr);
#ifdef VERBOSE
    printf("Allowing connections on socket [%d]\n", s);
#endif /* VERBOSE */
    /* s2 = accept (s, new_s_addr_ptr, &new_s_addr_len); */

    /* make sure the socket is a blocking socket i.e. hasn't been probe'd */
    setblocking (s);
    s2 = accept (s, &new_s_addr, &new_s_addr_len);

    if (s2<0) {
	printf("Accept on socket [%d] failed as [%d]\n", s, s2);
	perror ("Accept()");
#ifndef WIN32
        close (s2); /* not that it was open in the first place */
#else
        closesocket(s2);
#endif
	return (s2);
    }
#ifdef VERBOSE
    printf("Accepted connection on socket [%d] as new socket [%d]\n", s, s2);
#endif /* VERBOSE */

    nodelay(s2);

    return (s2);
}




/* Probe conn is a non blocking version of allowconn() */
int probeconn (s, fe, spid)
    int s;					/* socket we listen on */
int fe;					/* Fork Enabled? 0=no 1=yes */
/* If enabled then forks new process on accept */
int *spid;				/* new process id handling the returned socked */
/* used by multitasking servers */
{
    int s2;	/* new socket */
    /* struct sockaddr *new_s_addr_ptr;  */
    struct sockaddr new_s_addr; 
#ifndef WIN32
    size_t new_s_addr_len;
#else
    size_t new_s_addr_len;
#endif

    new_s_addr_len = sizeof (new_s_addr);
#ifdef VERBOSE
    printf("Allowing connections on socket [%d]\n", s);
#endif /* VERBOSE */
    /* s2 = accept (s, new_s_addr_ptr, &new_s_addr_len); */

    /* make the socket non blocking for the probe to work */
    setnonblocking (s);

    s2 = accept (s, &new_s_addr, &new_s_addr_len);
#ifdef VERBOSE
    printf("Accept returned [%d][0x%x]\n", s2, s2);
#endif /* VERBOSE */

    if (s2<0) {
	/* 	printf("Accept on socket [%d] failed as [%d]\n", s, s2); */
	/* 	printf("Errno is %d\n", errno); */
	/* 	printf("EWOULDBLOCK = %d\n", (int) EWOULDBLOCK); */
	/* as this is non-blocking printing the below is wrong as its not an */
	/* error, but an unavilable connection */
	/*  	perror ("Accept()"); */
	/* 	close (s2);  */
	/* not that it was open in the first place */
	/* be nice and make the return value 0 ? */
	/* 	s2 = 0; */
	/* 	return (s2); */
	return (0);
    }
#ifdef VERBOSE
    printf("Accepted connection on socket [%d] as new socket [%d]\n", s, s2);
#endif /* VERBOSE */

    /* make the new socket blocking by default as it is currently non */
    setblocking (s2);

    /* set tcpnodelay as well */
    nodelay(s2);

    return (s2);
}


int getconn (host, port, search)
    char *host;				/* full name of host running service */
int  *port;				/* port where service is running */
/* note this is also a return value */
int  search;			/* range allowed for port search */

{
    struct sockaddr_in  sa_in;
    struct hostent *hp;
    u_long  target_s_addr;
    int ts;
    int rc = 0;
#ifndef WIN32
    int i;
    unsigned short target_port;
#else
    int target_port;
#endif


    target_port = *port;					 /* get start point */

    hp = gethostbyname (host);
    if ( !hp ) {
	hp = gethostbyname("127.0.0.1");
	/*herror(host);*/
    }
    memcpy( &target_s_addr, hp->h_addr, hp->h_length );
    sa_in.sin_family       = AF_INET;
    sa_in.sin_addr.s_addr  = target_s_addr;

    while (search>=0) {
        ts = socket (AF_INET, SOCK_STREAM, 0);   /* yuck */
        /* try a new socket each time? */
        sa_in.sin_port         = htons((unsigned short)target_port);
        /* hope I can keep reusing this struct */

        rc = connect (ts, (struct sockaddr *) &sa_in, sizeof(sa_in));

        if (!rc) { /* success */
            *port = target_port;        /* tell them the port used */
#ifndef WIN32
            i = fcntl (ts, F_GETFL, 0);
#ifdef VERBOSE
            printf("FL on [%d] = %d\n", ts, i);
#endif /* VERBOSE */
#endif
            nodelay(ts);
            return (ts);            /* give them the socket */
        }
        else
            { /* failure */
                perror ("Connect()");
                target_port++;  /* look at the next port */
                search--;               /* one less attempt left */
                /* I belive ts is now defunted.. so I close it?? */
#ifndef WIN32
                close (ts);
#else
                closesocket( ts );
#endif
            }
    } /* while */

    /* well we are here so fidles.. */
    return (rc);
}


int getconn_addr (addr, port, search)
    unsigned long addr;		/* sa_addr of host running service */
int  *port;				/* port where service is running */
/* note this is also a return value */
int  search;			/* range allowed for port search */

{
    struct sockaddr_in  sa_in;
    int ts;
    int rc = 0;
#ifndef WIN32
    int i;
    int target_port;
#else
    unsigned short target_port;
#endif

    target_port = *port;					 /* get start point */

    /* hp = gethostbyname (host); */
    /* bcopy(hp->h_addr, (char *)&target_s_addr, hp->h_length); */
    sa_in.sin_family       = AF_INET;
    /* sa_in.sin_addr.s_addr  = target_s_addr; */
    sa_in.sin_addr.s_addr  = addr;

    while (search>=0) {
	ts = socket (AF_INET, SOCK_STREAM, 0);   /* yuck */
	/* try a new socket each time? */
	sa_in.sin_port         = htons (target_port); /* convert just to be safe */
	/* hope I can keep reusing this struct */

	rc = connect (ts, (struct sockaddr *) &sa_in, sizeof(sa_in));

	if (!rc) { /* success */
	    *port = target_port;	/* tell them the port used */
#ifndef WIN32
            i=fcntl (ts, F_GETFL, 0);
#ifdef VERBOSE
            printf("FL on [%d] = %d\n", ts, i);
#endif /* VERBOSE */
#endif
	    nodelay(ts);
	    return (ts);            /* give them the socket */
	}
	else
	    { /* failure */
		perror ("Connect()");
		target_port++;	/* look at the next port */
		search--;		/* one less attempt left */
		/* I belive ts is now defunted.. so I close it?? */
#ifndef WIN32
                close (ts);
#else
                closesocket(ts);
#endif
	    }
    } /* while */

    /* well we are here so fidles.. */
    return (rc);

}


int closeconn (s)
    int	s;					/* socket to close, also notifies the other party */

{
    /* too tired */
#ifndef WIN32
    close (s);
#else
    closesocket(s);
#endif
    return 0;
}



int writeconn (s, data, len)/* sends stream data. I.e. no header/seq stuff */	
    /* return value = len or error */
    int	s;					/* socket conn is on */
char *data;				/* raw data to send */
int len;				/* length of data to send */	

{
    int tosend, gone, fluffed;              /* loop variables */
    int i;


    tosend = len;
    gone = 0;       /* just in case */
    fluffed = 0;        /* just in case */

    for (i=0;tosend>0;) {
#ifdef DB9
	printf("Write loop for socket [%d]. Sent [%d] / [%d] or [%d] left.\n",
	       s, i, len, tosend);
#endif /* DB9 */

#ifndef WIN32
        gone = write (s, &data[i], tosend);
#else
        gone = send (s, (const char*)&data[i], tosend, 0 );
#endif

#ifdef DB9
	printf("wrote [%d] byte on socket [%d].\n", gone, s);
#endif /* DB9 */
	if (gone<=0) {
	    fluffed++;
	    if (gone<0) perror ("Write()");
	    if (fluffed==RETRIES) {
		fprintf(stderr,"Problem, connection on socket [%d] has failed.\nThis connect has been closed.\n", s); fflush(stderr);
#ifndef WIN32
                close (s);
#else
                closesocket(s);
#endif
		return (i); /* how much we sent in the end... */
		/* upto the user app to realise the error. */
	    } /* if fluffed */
	}
	else { /* i.e. we sent some */
	    i+= gone;   /* total sent and index ptr update */
	    tosend -= gone;
	    fluffed = 0;    /* fluffed reset counter */
	} /* if gone */
    } /* for look on tosend */


    /* ok all done... now return amount wrote */
    return (i);


}

int readconn (s, data, len)/* reads stream data. I.e. no header/seq stuff */	
    /* return is read length or error */
    int	s;					/* socket conn is on */
char *data;				/* raw data buffer */
int len;				/* length of data tobe read */

{
    unsigned int toget;  			/* note type */
    int got, fluffed;               /* loop variables */
    int i;


    toget = (unsigned) len;
    fluffed = 0;

#ifdef DB9
    printf("read on [%d] to get msg length [%d]\n", s, toget);
#endif

    for (i=0;toget>0;) {
#ifdef DB9
	printf("Reading on [%d] amount [%d]\n", s, toget);
#endif /* DB9 */

#ifndef WIN32
        got = read (s, &data[i], toget);
#else
        got = recv (s, (char *)&data[i],toget,0);
#endif

#ifdef DB9
	printf("reading on [%d] got [%d] bytes.\n", s, got);
#endif /* DB9 */

	if (got<=0) {
	    fluffed++;
	    if (fluffed==RETRIES) {
		fprintf(stderr,"Problem, connection on socket [%d] has failed.\nThis connect has been closed.\n", s); fflush(stderr);
#ifndef WIN32
                close (s);
#else
                closesocket(s);
#endif
		return (i); /* how much we got in the end... */
		/* upto the user app to realise the error. */
	    } /* if fluffed */
	}
	else { /* i.e. we got */
	    i+= got;    /* total got and index ptr update */
	    toget -= got;
	    fluffed = 0;    /* fluffed reset counter */
	} /* if got */
    } /* for look on toget */

    /* ok all done... now return amount read */
    return (i);



}


int writemsgconn (s, data, len)/* sends message data. I.e. header & seq */	
    /* return value = len or error */
    int	s;					/* socket conn is on */
char *data;				/* raw data to send */
int len;				/* length of data to send */	

{
    int tosend, gone, fluffed;				/* loop variables */
    int nblen;
    int i,n;

    nblen = htonl ((unsigned long) len);	/* message length for the header */
#ifdef DB9
    printf("Attempting message write on [%d] for msg of len [%d]\n", s, len);
#endif

    /* write outgoing message length */
#ifndef WIN32
    n = write (s, &nblen, sizeof(long));        /* ek */
#else
        n = send (s, (const char*) &nblen, sizeof(long), 0 );
#endif

    if(n!=4) { exit (-100); } /* ok thats a bad way to do it! */
#ifdef DB9
    printf("wrote on socket [%d] message hdr.\n", s);
#endif



    tosend = len;
    gone = 0; 		/* just in case */
    fluffed = 0;		/* just in case */

    for (i=0;tosend>0;) {
#ifdef DB9	
	printf("Write loop for socket [%d]. Sent [%d] / [%d] or [%d] left.\n", 
	       s, i, len, tosend);
#endif /* DB9 */

#ifndef WIN32
        gone = write (s, &data[i], tosend);
#else
        gone = send(s,(const char *)&data[i],tosend,0);
#endif


#ifdef DB9	
	printf("wrote [%d] byte on socket [%d].\n", gone, s);
#endif /* DB9 */

	if (gone<=0) {
	    fluffed++;
	    if (gone<0) perror ("Write()");
	    if (fluffed==RETRIES) {
		fprintf(stderr,"Problem, connection on socket [%d] has failed.\nThis connect has been closed.\n", s); fflush(stderr);
#ifndef WIN32
                close (s);
#else
                closesocket(s);
#endif
		return (i);	/* how much we sent in the end... */
		/* upto the user app to realise the error. */
	    } /* if fluffed */
	}
	else { /* i.e. we sent some */
	    i+= gone;	/* total sent and index ptr update */
	    tosend -= gone;
	    fluffed = 0;	/* fluffed reset counter */
	} /* if gone */
    } /* for look on tosend */


    /* ok all done... now return amount wrote */
    return (i);

}


int readmsgconn (s, data, mlen)/* reads message data. I.e. header & seq */	
    /* return value = len or error */
    int	s;					/* socket conn is on */
char *data;				/* raw data buffer */
unsigned int mlen;				/* max length of data buffer */	
{
    int i,k;
    unsigned int j;
    int n;
    char bb;						/* Bit Bucket character */
    unsigned long ilen, nbilen;		/* incomming message length */
    unsigned int	tojunk, toget;	/* note type */
    int got, fluffed;				/* loop variables */

    fluffed = 0;


    /* read incomming message length */
#ifndef WIN32
    n = read (s, &nbilen, sizeof(long));        /* ek */
#else
        n = recv (s, (char *)&nbilen, sizeof(long),0);
#endif

    if(n!=4) { exit (-100); } /* ok thats a bad way to do it! */

    ilen = ntohl (nbilen);	/* convert from network byte format */

#ifdef DB9
    printf("read on socket [%d] returns message hdr indicating msg size of [%u]\n",
	   s, ilen);
#endif

    if (ilen > mlen)  	/* buffer too small.. will have to truncate. */
	/* Note we don't read ahead or anything fancy */
	{
	    tojunk = ilen - mlen;
	    toget  = mlen;
	}
    else				{
	tojunk = 0;
	toget = ilen;
    }
#ifdef DB9
    printf("read on [%d] total msg length [%d] overrun [%d]\n", s, toget, tojunk);
#endif

    for (i=0;toget>0;) {
#ifdef DB9	
	printf("Reading on [%d] amount [%d]\n", s, toget);
#endif /* DB9 */

#ifndef WIN32
        got = read (s, &data[i], toget);
#else
        got = recv(s, (char *)&data[i],toget,0);
#endif

#ifdef DB9	
	printf("reading on [%d] got [%d] bytes.\n", s, got);
#endif /* DB9 */

	if (got<=0) {
	    fluffed++;
	    if (fluffed==RETRIES) {
		fprintf(stderr,"Problem, connection on socket [%d] has failed.\nThis connect has been closed.\n", s); fflush(stderr);
#ifndef WIN32
                close (s);
#else
                closesocket(s);
#endif
		return (i);	/* how much we got in the end... */
		/* upto the user app to realise the error. */
	    } /* if fluffed */
	}
	else { /* i.e. we got */
	    i+= got;	/* total got and index ptr update */
	    toget -= got;
	    fluffed = 0;	/* fluffed reset counter */
	} /* if got */
    } /* for look on toget */

    /* now for the truncate bit. */
    if (tojunk)
	for (j=0;j<tojunk;) {
#ifndef WIN32
            k = read (s, &bb, 1);       /* yunk one at a time... */
#else
                k = recv(s,(char*)&bb,1,0);
#endif
	    if (k==1) j++;			/* or j+= k ? */
	    /* should check here for k failing.. but hey its late. */
	}

    /* ok all done... now return amount read */
    return (i);
}

/* this routine sets the socket to no delay */
void nodelay( s)
    int s;
{
    int one;
    struct protoent *p;
    p = getprotobyname("tcp");
    if( p && setsockopt(s, p->p_proto, TCP_NODELAY, (char*)&one, sizeof(one)) < 0);
    if( p && setsockopt(s, p->p_proto, SO_REUSEADDR, (char*)&one, sizeof(one)) < 0)
	perror("setsockopt: resuseaddr");
}

void setnonblocking( s ) /* sets the given socket non-blocking */
    int s;
{
    int i,j;
#ifndef WIN32
    i = fcntl (s, F_GETFL, 0);
#ifdef VERBOSE
    printf("FL on [%d] = %d\n", s, i);
#endif /* VERBOSE */
    i = i | O_NONBLOCK; /* mark it as non blocking */
#ifdef VERBOSE
    printf ("O_NONBLOCK = %d\n", O_NONBLOCK);
#endif /* VERBOSE */
    j = fcntl (s, F_SETFL, i);
#ifdef VERBOSE
    printf("New FL [%d] on [%d] returns [%d]\n", i, s, j);
#endif /* VERBOSE */
#else
        j = 1;
        i = ioctlsocket( s, FIONBIO, &j);
#endif
}

void setblocking( s ) /* sets the given socket blocking again */
    int s;
{
    int i,j;
#ifndef WIN32
    i = fcntl (s, F_GETFL, 0);
#ifdef VERBOSE
    printf("FL on [%d] = %d\n", s, i);
#endif /* VERBOSE */
    i = i & (!O_NONBLOCK); /* mark it as non non-blocking, i.e. blocking */
    j = fcntl (s, F_SETFL, i);
#ifdef VERBOSE
    printf("New FL [%d] on [%d] returns [%d]\n", i, s, j);
#endif /* VERBOSE */
#else
        j = 0;
        i = ioctlsocket( s, FIONBIO, &j);
#endif
}


int tcp_set_option( int fd, int type, int kind, int desired,
	char *string )
{
   int result, option;
#ifndef WIN32
   size_t optlen = sizeof(option);
#else
   int optlen = sizeof(option);
#endif

   if ( getsockopt(fd, type, kind, (char *)&option, &optlen ) == -1 ) {
        fprintf(stderr,"tcp_set_option(): getsockopt(%s): ",string);
        perror(NULL);
        return(-1);
   }

#ifdef DEBUG
   fprintf(stdout, "%s is %d, length %d\n",string,option,optlen);
#endif

   if (desired<0) return(option);

   if (desired<option) {
#ifdef DEBUG
      fprintf(stdout, "%s attempting %d, length %d\n",string,desired,
        sizeof(option));
#endif
      if(setsockopt(fd, type, kind, (char *)&desired,sizeof(option)) == -1 ){
        fprintf(stderr,"tcp_set_option(): setsockopt(%s): ", string);
        perror(NULL);
        return(option);
      }
   }

   else while (1) {
        option+=(((desired-option)>64)?64:(desired-option));
        if ( setsockopt(fd,type,kind,(char*)&result,optlen)==-1)
            break;
        if( getsockopt(fd,type,kind,(char*)&result,&optlen)==-1)
           break;
        if( (result<option) || (result>=desired))
           break;
   }

   if (getsockopt(fd,type,kind,(char*)&option,&optlen)==-1) {
        fprintf(stderr,"tcp_setoption(): getsockopt(%s): ", string);
        perror(NULL);
        return(-1);
   }
#ifdef DEBUG
   fprintf(stdout,"%s is now %d, length %d\n", string,option,optlen);
#endif
   return option;
}


int setsendrecvbufs ( s, bufsize ) /* Sets send/recv buffer sizes */
    /* returns 0 for ok and 1 for unix not accepting and -1 for error */
    /* if a problem occurs, it sets buffers back to what they were */
    /* Note: Bufsize is in kilobytes */
{
    unsigned int sb_org, rb_org;
    unsigned int sb_new, rb_new;
    unsigned int sb_chk, rb_chk;
#ifndef WIN32
    size_t optsize;
#else
    int optsize;
#endif

    optsize = sizeof (int); /* I hope... */

    sb_new = bufsize * 1024;	/* into bytes */
    rb_new = bufsize * 1024;	/* into bytes */

    if (sb_new<=0) {
	fprintf(stderr,
		"SNIPE_LITE:Conn:attempt to set send/recv bufs to %d ignored\n",
		sb_new);
	return (-1);
    }
	
    /* get original values first */

    getsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&sb_org, &optsize);		
    getsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&rb_org, &optsize);  

    /* Now to attempt to set the new ones */

    setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&sb_new, optsize);
    setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&rb_new, optsize);

    /* Now to verify the socket options */

    getsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&sb_chk, &optsize);		
    getsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&rb_chk, &optsize);  

    /* ok junk debugging */
#ifdef VERBOSE
    printf("[sockopt] socket %d send org %d new %d vrfy %d\t\trecv org %d, new %d vrfy %d\n",
	   s, sb_org, sb_new, sb_chk, rb_org, rb_new, rb_chk);
#endif

    /* Now to do what we do */

    /* if both are ok... worked */
    if ((sb_chk==sb_new)&&(rb_chk==rb_new)) return (0);

    /* if **either** match the older values */
    if ((sb_chk==sb_org)||(rb_chk==rb_org)) return (1);

    /* else we have to attempt a reset */
    setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&sb_org, optsize);
    setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&rb_org, optsize);

    /* At this point we could check again, but there is little point as */
    /* if this hasn't changed it back we are in trouble anyway */
    return (1);
}

double sec_time() 
    /* returns the time in seconds as a double */
{
    double sec=0.0;
    double psec=0.0;
#ifndef WIN32
    struct timeval tp;
    struct timezone tzp;

    gettimeofday (&tp, &tzp);
    sec = (double)tp.tv_sec;
    psec = ((double)tp.tv_usec)/1000000.0;
#else
        struct tm *today;
    struct _timeb tstruct;

        _ftime( &tstruct );
        today = localtime( (const time_t *) &tstruct.time );
        sec = today->tm_sec;
        psec = (((double)tstruct.millitm)/1000000.0);
#endif
    return (sec+psec);
}
