/*
$Author: mucci $
$Date: 2002/08/10 10:42:18 $
$Revision: 1.12 $
$Source: /cvs/homes/papi/papi/tools/perfometer/server.c,v $
*/
/*************************************************************************
Created by:	Thomas Spencer		spencer@cs.utk.edu
Modified by:	Kevin London		london@cs.utk.edu
File:	server.c
Date:	3/2001
Description:	Relays and processes messages between a PERF
		GUI and PERF applications.
Usage:	server [hostname] [port] [hostname] [port] ...
	Applications can be specified on the cmd line or
	received from the GUI but not both.
Option:	-f [filename]
	All messages received are written to output file.
Option:	-nogui
	For testing without gui.
*************************************************************************/

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sched.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <signal.h>
#include "conn.h"
#include "perfometer_server.h"

#ifdef SOLARIS
  #include <sys/file.h>
#endif

#define PERF_BREACH PERF_BREECH

#define GUI_TO_PORTNUM 3545
#define GUI_FROM_PORTNUM 7879
#define REC_BUF_LIMIT 10
#define REC_BUF_RESTART 5
#define SELECT_TIME_LIMIT 10
#define DEFAULT_EVENTS "PAPI_TOT_CYC@PAPI_FP_INS"

#define for_apps(ptr) \
  for(ptr = data->app_list->next; ptr != data->app_list; ptr = ptr->next)

struct Data{
  int num_apps;
  struct App *gui;
  struct App *app_list;
  fd_set masterfds;
  pthread_mutex_t rec_mutex;
};

struct App{
  int rank;
  int port_to;
  int port_from;
  char hostname[256];
  char ip[24];
  int fd_to;
  int fd_from;
  int num_records;
  int rec_limit_reached;
  struct Record *rec_list;
  struct Record *rec_list_end;
  struct Threshhold *thhd_list;
  struct App *next;
  struct App *prev;
};

struct Record{
  int type;
  char *message;
  struct Record *next;
};

struct Threshhold{
  float max;
  float min;
  int breach;
  int data_num;
  int variance;
  int var_count;
  struct Threshhold *prev;
  struct Threshhold *next;
};

int init_server(int, char**);
struct App *init_app(char*, char*);
int auth_gui();
int auth_app(struct App*);
int run_threads();
int read_message(int, int*, char*);
int write_message(int, int, char*);

void* gui_to_comm();
static void gui_to_app_comm();
void* app_from_comm();

int create_record(struct App*);
int send_record(struct App*, struct App*);
int delete_record(struct App*);
int add_app(struct App*);
int delete_app(struct App*);

static struct Data *data;
FILE *output_file = NULL;
char *event_list = NULL;
int GUI = 1;
int GUI_fd = -1;

int main(int argc, char **argv)
{
  int i, rc, events = 0;
  struct App *temp_app;

  system("clear");

  for(i = 1; i < argc; i++){
    rc = strcmp(argv[i], "-h");
    if(rc == 0){
      fprintf(stdout,
"ARGUMENTS AND OPTIONS:\nserver [-nogui eventname@eventname@...] [-f filename] host port ...\n* -nogui eventname      This will allow applications to be monitored\n                        without the GUI. Multiple events to be monitored can\n                        be specified, separated by @. Eventnames are\n                        recognized by beginning with PAPI_. If something else\n                        is given for eventname, the results are unspecified.\n                        If no eventnames are specified, the default of \n                        PAPI_TOT_CYC@PAPI_FP_INS will be used.\n* -f filename           This creates a trace file that the GUI\n                        can later read and use to graph the results.\n* host port             These are the host and port number of the\n                        application[s] that are to be monitored.\n* -h                    Short description on how to use the server.\n");
      fprintf(stdout,
"\nTO RUN:\nFirst, start all of the applications that are to be monitored. They will display the port number that they are listening on. For an MPI application, only one port will be provided, the rest will be auto-exchanged. There are two options, you can specify the applications on the server command line, or you can specify the applications in the GUI. Second, execute the server and it will print out a port number that it is listening on. Third, start the GUI and connect to the host and port the server is on. If the applications are supplied on the server's command line, then all the connections will be setup and the GUI will begin graphing the performance data from all the applications. If the applications are NOT supplied by the server, a pop-up window will appear in the GUI. The information is then typed into the list, with one host and port per line, such as:\n     somewhere.com 3545\n     another.com 3545\nThen, all the connections will be setup and the GUI will begin graphing the performance data.\n");
      exit(1);
    }
  }

  #ifdef DEBUG
    fprintf(stdout, "Server starting...\n");
  #endif

  for(i = 1; i < argc; i++){
    rc = strcmp(argv[i], "-nogui");
    if(rc == 0){
      GUI = 0;
      if(argc >= i+2){
        rc = strncmp("PAPI_", argv[i+1], sizeof(char)*5);
        if(rc == 0){
          event_list = strdup(argv[i+1]);
          if(event_list == NULL){
            perror("Malloc");
            exit(-1);
          }
          else{
            events = 1;
            i++;
            continue;
          }
        }
      }
      event_list = strdup(DEFAULT_EVENTS);
      if(event_list == NULL){
        perror("Malloc");
        exit(-1);
      }
      else
        events = 0;
      continue;
    }
    rc = strcmp(argv[i], "-f");
    if(rc == 0){
      if(argc < i+2){
        fprintf(stderr,
          "USAGE: server [-h] [-nogui] [-f output_file] [host] [port] ...\n");
        exit(-1);
      }
      output_file = fopen(argv[i+1], "w");
      if(output_file == NULL){
        fprintf(stderr, "Cannot open output file %s.\n", argv[i+1]);
        exit(-1);
      }
      else{
        i++;
        continue;
      }
    }
  }

  i = 0;
  if(GUI == 0)
    i++;
  if(events == 1)
    i++;
  if(output_file != NULL)
    i += 2;

  argc = argc - i;
  argv = &argv[i];

  /* Set up structures for server */
  rc = init_server(argc, argv);
  if(rc != 1){
    fprintf(stderr, "Init_server failed.\n");
    fprintf(stderr, "Exiting server.\n");
    return(-1);
  }

  /* Create and authorize connections to and from GUI */
  rc = auth_gui();
  if(rc != 1){
    fprintf(stderr, "Auth_gui failed.\n");
    fprintf(stderr, "GUI connection not authorized.\n");
    fprintf(stderr, "Exiting server.\n");
    return(-1);
  }

  #ifdef DEBUG
    fprintf(stdout, "Establishing Application connections.\n");
  #endif

  /* Connect and authorize connections to and from Applications */
  for_apps(temp_app){
    rc = auth_app(temp_app);
    if(rc != 1){
      fprintf(stderr, "Application connection not authorized for ");
      fprintf(stderr, "%s %d.\n", temp_app->hostname, temp_app->port_from);
      rc = write_message(data->gui->fd_to, PERF_FAILEDCONN, "");
      if(rc != 1){
        fprintf(stderr, "Write_message failed.\n");
        fprintf(stderr, "Error writing PERF_FAILEDCONN to GUI.\n");
      }
      return(-1);
    }

    #ifdef DEBUG
      fprintf(stdout, "Connection made from %s %d on socket %d.\n",
        temp_app->hostname, temp_app->port_from, temp_app->fd_from);
      fprintf(stdout, "Connection made to %s %d on socket %d.\n",
        temp_app->hostname, temp_app->port_to, temp_app->fd_to);
    #endif
  }

  /* Send message to GUI that all connections were successful */
  rc = write_message(data->gui->fd_to, PERF_ACK, "");
  if(rc != 1){
    fprintf(stderr, "Write_message failed. Retrying.\n");
    rc = write_message(data->gui->fd_to, PERF_ACK, "");
    if(rc != 1){
      fprintf(stderr, "Write_message failed.\n");
      fprintf(stderr, "Exiting server.\n");
      return(-1);
    }
  }

  #ifdef DEBUG
    fprintf(stdout, "All connections made.\n");
  #endif

  /* Launch processes for passing data between GUI and Apps */
  rc = run_threads();
  if(rc != 1){
    fprintf(stderr, "Run_threads failed.\n");
    fprintf(stderr, "Exiting server.\n");
    return(-1);
  }

  pthread_exit(NULL);
  return(1);
}

/*------------------------------------------------------------------------
Function name:	init_server
Input parameters:	int argc, char **argv
Return value:	-1 if error, 1 if successful
Description:	Creates initial memory and sets initial values
		for shared data structure.
------------------------------------------------------------------------*/

int init_server(int argc, char **argv){

  int i, rc;
  struct App *new_app;

  #ifdef DEBUG
    fprintf(stdout, "Setting up data structure.\n");
  #endif

  /* Create data structure and initialize all values */
  data = (struct Data*)malloc(sizeof(struct Data));
  if(data == NULL){
    perror("Malloc");
    return(-1);
  }
  data->num_apps = 0;
  data->app_list = NULL;
  FD_ZERO(&data->masterfds);
  pthread_mutex_init(&data->rec_mutex, NULL);

  /* Create list for App structures */
  data->app_list = (struct App*)malloc(sizeof(struct App));
  if(data->app_list == NULL){
    perror("Malloc");
    return(-1);
  }

  data->app_list->rank = 0;
  data->app_list->port_from = 0;
  data->app_list->port_to = 0;
  data->app_list->hostname[0] = '\0';
  data->app_list->ip[0] = '\0';
  data->app_list->fd_to = -1;
  data->app_list->fd_from = -1;
  data->app_list->num_records = 0;
  data->app_list->rec_limit_reached = 0;
  data->app_list->rec_list = NULL;
  data->app_list->rec_list_end = NULL;
  data->app_list->thhd_list = NULL;
  data->app_list->prev = data->app_list;
  data->app_list->next = data->app_list;

  data->gui = data->app_list;

  /* Create App data structures if given on command line */
  for(i=0; i < (argc-1)/2; i++){
    new_app = init_app(argv[i*2+1], argv[i*2+2]);
    if(new_app == NULL){
      fprintf(stderr, "Init_app failed.\n");
      return(-1);
    }
    rc = add_app(new_app);
    if(rc != 1){
      fprintf(stderr, "Add_app failed.\n");
      return(-1);
    }
  }

  #ifdef DEBUG
    fprintf(stdout, "Data structure setup is done.\n");
  #endif

  return(1);
}

/*------------------------------------------------------------------------
Function name:	init_app
Input parameters:	char *host, char *port
Return value:	NULL if error, struct App *new_app if successful
Description:	Initialize new application.
------------------------------------------------------------------------*/

struct App* init_app(char *host, char *port){

  struct App *new_app;

  #ifdef DEBUG
    fprintf(stdout, "Setting up Application structure.\n");
  #endif

  new_app = (struct App*)malloc(sizeof(struct App));
  if(new_app == NULL){
    perror("Malloc");
    return(NULL);
  }

  /* Store values given in each App */
  strcpy(new_app->hostname, host);

  if(sscanf(port, "%d", &new_app->port_from) != 1){
    fprintf(stderr, "ERROR: Port %s given for ", port);
    fprintf(stderr, "host %s is not a number.\n", host);
    return(NULL);
  }
  if(new_app->port_from <= 0){
    fprintf(stderr, "ERROR: Port must be > 0.\n");
    fprintf(stderr, "For host %s, port %s was given.\n", host, port);
    return(NULL);
  }

  new_app->rank = data->app_list->prev->rank + 1;
  data->app_list->port_to = 0;
  new_app->ip[0] = '\0';
  new_app->fd_to = -1;
  new_app->fd_from = -1;
  new_app->num_records = 0;
  new_app->rec_limit_reached = 0;
  new_app->rec_list = NULL;
  new_app->rec_list_end = NULL;
  new_app->thhd_list = NULL;

  return(new_app);
}

/*------------------------------------------------------------------------
Function name:	auth_gui
Input parameters:	None
Return value:	-1 if error, 1 if successful
Description:	Tests whether the connection is from an
		appropriate perf gui.
------------------------------------------------------------------------*/

int auth_gui(){

  int rc, i;
  int type;
  int length, size;
  char *mp;
  char port_num[64];
  char message[MAX_BUFFER_SIZE];
  char input[MAX_BUFFER_SIZE];
  struct hostent *hp;
  struct in_addr in;
  struct App *temp_app;

  #ifdef DEBUG
    fprintf(stdout, "Establishing GUI connection.\n");
  #endif

  if(GUI){
    /* Create socket for GUI to connect to */
    rc = setportconn(&data->gui->fd_to, GUI_TO_PORTNUM, 0);
    if(rc <= 0){
      fprintf(stderr, "Setportconn failed.\n");
      return(-1);
    }
    else data->gui->port_to = rc;
    fprintf(stderr, "********************************\n");
    fprintf(stderr, "*** OPENED SOCKET ON PORT: %d ***\n", data->gui->port_to);
    fprintf(stderr, "********************************\n");
    rc = allowconn(data->gui->fd_to, 0, 0);
    if(rc <= 0){
      fprintf(stderr, "Allowconn failed.\n");
      return(-1);
    }
    else{
      data->gui->fd_to = rc;
    }
    #ifdef DEBUG
      fprintf(stdout, "Connection made to GUI on socket %d and port %d.\n",
        data->gui->fd_to, data->gui->port_to);
    #endif
  }

  GUI_fd = data->gui->fd_to;

  #ifdef DEBUG
    fprintf(stdout, "GUI connection established.\n");
    fprintf(stdout, "Verifying connection to GUI.\n");
  #endif

  /* Create string with server info to send to GUI */
  rc = gethostname(input, MAX_BUFFER_SIZE);
  if(rc != 0 || input[0] == '\0'){
    perror("Gethostname");
    return(-1);
  }
  hp = gethostbyname(input);
  if(hp == NULL){
    perror("Gethostbyname");
    return(-1);
  }

  memcpy(&in.s_addr, hp->h_addr_list[0], sizeof(in.s_addr));
  cuserid(input);
  if(input[0] == '\0'){
    perror("Cuserid");
    return(-1);
  }
  if(data->num_apps < 1){
    sprintf(message, "%s%s%s%s%s%d", PERF_DELIM, input, PERF_DELIM,
      inet_ntoa(in), PERF_DELIM, PERF_NEEDAPP);
  }
  else{
    sprintf(message, "%s%s%s%s%s%d%s%d", PERF_DELIM, input,
      PERF_DELIM, inet_ntoa(in), PERF_DELIM,
      PERF_NONEEDAPP, PERF_DELIM, data->num_apps);
  }
  rc = write_message(data->gui->fd_to, PERF_SERVER, message);
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }
  strcpy(message, "");

  if(GUI){
    /* Read a PERF_ACK from GUI to confirm message was received */
    rc = read_message(data->gui->fd_to, &type, message);
    if(rc <= 0){
      fprintf(stderr, "Read_message failed.\n");
      return(-1);
    }
    if(type != PERF_ACK){
      fprintf(stderr, "PERF_ACK was not received.\n");
      return(-1);
    }
    strcpy(message, "");

    #ifdef DEBUG
      fprintf(stdout, "Verified connection to GUI.\n");
      fprintf(stdout, "Creating connection from GUI.\n");
    #endif

    /* Create a new socket for second connection to GUI */
    rc = setportconn(&data->gui->fd_from, GUI_FROM_PORTNUM, 0);
    if(rc <= 0){
      fprintf(stderr, "Setportconn failed.\n");
      return(-1);
    }
    else data->gui->port_from = rc;
  }

  /* Send GUI new port to connect on */
  rc = write_message(data->gui->fd_to, data->gui->port_from, "");
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }

  if(GUI){
    /* Create connection to GUI on new socket */
    rc = allowconn(data->gui->fd_from, 0, 0);
    if(rc < 0){
      fprintf(stderr, "Allowconn failed.\n");
      return(-1);
    }
    else{
      data->gui->fd_from = rc;
    }

    #ifdef DEBUG
      fprintf(stdout, "Connection made from GUI on socket %d and port %d.\n",
        data->gui->fd_from, data->gui->port_from);
    #endif

    /* Read a PERF_ACK from GUI to confirm connection */
    rc = read_message(data->gui->fd_from, &type, message);
    if(rc <= 0){
      fprintf(stderr, "Read_message failed.\n");
      return(-1);
    }
    if(type != PERF_ACK){
      fprintf(stderr, "PERF_ACK was not received.\n");
      return(-1);
    }
  }

  /* Send PERF_ACK on new connection */
  rc = write_message(data->gui->fd_to, PERF_ACK, "");
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }

  /* If Apps given on cmd line, send info to GUI */
  if(data->num_apps > 0){
    for_apps(temp_app){
      sprintf(message, "%s%s%s%d%s%d", PERF_DELIM, temp_app->hostname,
        PERF_DELIM, temp_app->port_from, PERF_DELIM, temp_app->rank);
      rc = write_message(data->gui->fd_to, PERF_SERVER, message);
      if(rc != 1){
        fprintf(stderr, "Write_message failed.\n");
        return(-1);
      }
      strcpy(message, "");
    }
    rc = write_message(data->gui->fd_to, PERF_END, "");
    if(rc != 1){
      fprintf(stderr, "Write_message failed.\n");
      return(-1);
    }
    strcpy(message, "");
  }

  /* If Apps not given on cmd line then receive info from GUI */
  else if(data->num_apps < 1 && GUI){
    while(1){
      /* Read App info from GUI */
      rc = read_message(data->gui->fd_from, &type, message);
      if(rc <= 0){
        fprintf(stderr, "Read_message failed.\n");
        return(-1);
      }
      if(type != PERF_APP && type != PERF_END){
        fprintf(stderr, "PERF_APP was not received.\n");
        return(-1);
      }
      if(type == PERF_END){
        break;
      }

      /* Parse and store Application info */
      i = 1;
      size = strlen(message);
      for(mp=(message+1); mp<=(message+size); mp+=(length+1), i++){
        length = strcspn(mp, PERF_DELIM);
        if(i == 1){
          strncpy(input, mp, length);
          input[length] = '\0';
        }
        if(i == 2){
          strncpy(port_num, mp, length);
          port_num[length] = '\0';
          break;
        }
      }

      /* Create application data structure */
      temp_app = init_app(input, port_num);
      if(temp_app == NULL){
        fprintf(stderr, "Init_app failed.\n");
        return(-1);
      }
      rc = add_app(temp_app);
      if(rc != 1){
        fprintf(stderr, "Add_app failed.\n");
        return(-1);
      }
    }
  }

  /* Error check if the App info was received */
  if(data->num_apps < 1){
    fprintf(stderr, "No applications were specified.\n");
    return(-1);
  }

  #ifdef DEBUG
    fprintf(stdout, "Connection to GUI authorized.\n");
  #endif

  return(1);
}

/*------------------------------------------------------------------------
Function name:	auth_app
Input parameters:	struct App *app
Return value:	-1 if error, 1 if successful
Description:	Tests whether the connection is from an
		appropriate perf application.
------------------------------------------------------------------------*/

int auth_app(struct App *app){

  int rc, i;
  int type;
  char message[MAX_BUFFER_SIZE];
  int length, size;
  char *mp;
  char input[256];
  char host[256], port_num[64];
  struct App *new_app;

  #ifdef DEBUG
    fprintf(stdout, "Establishing connection with app %d.\n", app->rank);
  #endif

  /* Connect to socket opened from Applications */
  rc = getconn(app->hostname, &app->port_from, 0);
  if(rc <= 0){
    fprintf(stderr, "Getconn failed.\n");
    return(-1);
  }
  else{
    app->fd_from = rc;
  }

  #ifdef DEBUG
    fprintf(stdout, "Verifying connection with App %d.\n", app->rank);
  #endif

  /* Read Application info */
  rc = read_message(app->fd_from, &type, message);
  if(rc <= 0){
    fprintf(stderr, "Read_message failed.\n");
    return(-1);
  }
  if(type != PERF_APP){
    fprintf(stderr, "Type PERF_APP was not received.\n");
    return(-1);
  }

  /* Parse and store Application info */
  i = 1;
  size = strlen(message);
  for(mp=(message+1); mp<=(message+size); mp+=(length+1), i++){
    length = strcspn(mp, PERF_DELIM);
    strncpy(input, mp, length);
    input[length] = '\0';

    /* Will be used in the future to authenticate user login */
    /* if(i==1){ */
    /*   if(strcmp(input,PERF_ANY) != 0 && */
    /*     strcmp(input,cuserid(uid))!=0){ */
    /*     fprintf(stderr, "User is not authorized.\n"); */
    /*     return(-1); */
    /*   } */
    /* } */

    /* Store IP address of Application */
    if(i==2){
      strcpy(app->ip, input);
      if(strlen(app->ip) == 0){
        fprintf(stderr, "IP address received invalid.\n");
        return(-1);
      }
    }
    else if(i==3){}
    else if(i==4){}

    /* Error check for extra pieces of data on message received */
    else if (i >= 5){
      fprintf(stderr, "Message was not properly formed.\n");
      return(-1);
    }
  }

  /* Send PERF_ACK to Application */
  sprintf(message, "%s%d", PERF_DELIM, app->rank);
  rc = write_message(app->fd_from, PERF_ACK, message);
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }
  strcpy(message, "");

  /* Receive new port to make second connection on */
  rc = read_message(app->fd_from, &type, message);
  if(rc <= 0){
    fprintf(stderr, "Read_message failed.\n");
    return(-1);
  }
  if(type <= 0){
    fprintf(stderr, "Valid port number was not received.\n");
    fprintf(stderr, "Port must be > 0.\n");
    return(-1);
  }
  app->port_to = type;

  /* Make second connection to application */
  rc = getconn(app->hostname, &app->port_to, 0);
  if(rc <= 0){
    fprintf(stderr, "Getconn failed.\n");
    return(-1);
  }
  else app->fd_to = rc;

  /* Send PERF_ACK to application */
  rc = write_message(app->fd_to, PERF_ACK, "");
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }
  strcpy(message, "");

  /* Receive PERF_ACK from application */
  rc = read_message(app->fd_from, &type, message);
  if(rc <= 0){
    fprintf(stderr, "Read_message failed.\n");
    return(-1);
  }
  if(type != PERF_ACK){
    fprintf(stderr, "Type PERF_ACK was not received.\n");
    return(-1);
  }

  if(strlen(message) > 0){
    /* Parse and store application info */
    i = 1;
    size = strlen(message);
    for(mp=(message+1); mp<=(message+size); mp+=(length+1), i++){
      length = strcspn(mp, PERF_DELIM);
      strncpy(input, mp, length);
      input[length] = '\0';

      /* Store Application hostname */
      if(i%2 == 1){
        strcpy(host, input);
      }
      if(i%2 == 0){
        strcpy(port_num, input);

        /* Create application data structure */
        new_app = init_app(host, port_num);
        if(new_app == NULL){
          fprintf(stderr, "Init_app failed.\n");
          return(-1);
        }
        rc = add_app(new_app);
        if(rc != 1){
          fprintf(stderr, "Add_app failed.\n");
          return(-1);
        }
      }
    }
  }
  strcpy(message, "");

  /* Pass hwinfo from Application to GUI */
  for(i = 0; i < 3; i++){
    rc = read_message(app->fd_from, &type, message);
    if(rc <= 0){
      fprintf(stderr, "Read_message failed.\n");
      return(-1);
    }
    rc = write_message(data->gui->fd_to, type, message);
    if(rc != 1){
      fprintf(stderr, "Write_message failed.\n");
      return(-1);
    }
    strcpy(message, "");
  }

  /* Receive PERF_ACK from GUI */
  rc = read_message(data->gui->fd_from, &type, message);
  if(rc <= 0){
    fprintf(stderr, "Read_message failed.\n");
    return(-1);
  }
  if(type != PERF_ACK){
    fprintf(stderr, "Type PERF_ACK was not received.\n");
    return(-1);
  }
  strcpy(message, "");

  /* Send PERF_ACK to Application */
  rc = write_message(app->fd_to, PERF_ACK, "");
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }
  strcpy(message, "");

  /* Set fd in list for reading from */
  FD_SET(app->fd_from, &data->masterfds);

  #ifdef DEBUG
    fprintf(stdout, "Connection with App %d verified.\n", app->rank);
  #endif

  return(1);
}

/*------------------------------------------------------------------------
Function name:	run_threads
Input parameters:	None
Return value:	-1 if error, 1 if successful
Description:	Creates threads for sending and receiving data.
------------------------------------------------------------------------*/

int run_threads(){

  int rc;
  char message[MAX_BUFFER_SIZE];
  struct App *temp_app;
  pthread_t *tcb_gui_to, *tcb_app_from;
  pthread_attr_t *custom_attr;
  /* struct sched_param fifo_param; */

  #ifdef DEBUG
    fprintf(stdout, "Creating threads.\n");
  #endif

  tcb_gui_to = (pthread_t*)malloc(sizeof(pthread_t));
  if(tcb_gui_to == NULL){
    perror("Malloc");
    return(-1);
  }
  tcb_app_from = (pthread_t*)malloc(sizeof(pthread_t));
  if(tcb_app_from == NULL){
    perror("Malloc");
    return(-1);
  }
  custom_attr = (pthread_attr_t*)malloc(sizeof(pthread_attr_t));
  if(custom_attr == NULL){
    perror("Malloc");
    return(-1);
  }
  pthread_attr_init(custom_attr);
  pthread_attr_setdetachstate(custom_attr, PTHREAD_CREATE_DETACHED);

  #ifdef _POSIX_THREAD_PRIORITY_SCHEDULING
    fprintf(stdout, "Setting thread priority schedule.\n");
    pthread_attr_setscope(custom_attr, PTHREAD_SCOPE_SYSTEM);

    /* May be used later to increase performance */
    /* fifo_param.sched_priority = */
    /*   (sched_get_priority_max(SCHED_FIFO) + */
    /*   sched_get_priority_min(SCHED_FIFO))/2 */
    /* pthread_attr_setschedparam(custom_attr, &fifo_param); */
  #endif

  #ifdef DEBUG
    fprintf(stdout, "Threads created.\n");
    fprintf(stdout, "Executing threads.\n");
  #endif

  if(pthread_create(tcb_gui_to, custom_attr, gui_to_comm, NULL) != 0){
    perror("Pthread_create");
    return(-1);
  }
  if(pthread_create(tcb_app_from, custom_attr, app_from_comm, NULL) != 0){
    perror("Pthread_create");
    return(-1);
  }

  #ifdef DEBUG
    fprintf(stdout, "Threads executed.\n");
  #endif

  if(!GUI){
    for_apps(temp_app){
      sprintf(message, "%s%d%s%s", PERF_DELIM, temp_app->rank,
        PERF_DELIM, event_list);
      rc = write_message(temp_app->fd_to, PERF_CHANGE_EVENT, message);
      if(rc != 1){
        fprintf(stderr, "Write_message failed.\n");
        return(-1);
      }
    }
  }

  return(1);
}

/*------------------------------------------------------------------------
Function name:	read_message
Input parameters:	int fd, int *type, char *message
Return value:	0 or -1 if error, 1 if successful
Description:	Reads a message from fd.
------------------------------------------------------------------------*/

int read_message(int fd, int *type, char *message){

  int rc;
  int length, read_length;

  if( !( (GUI_fd==fd) && !GUI) ){
    rc = read(fd, (char*)&length, sizeof(int));
    if(rc <= 0){
      perror("Read");
      return(rc);
    }
    if(rc != sizeof(int)){
      return(-1);
    }
    length = ntohl(length);
    if(length < 0){
      return(-1);
    }

    rc = read(fd, (char*)type, sizeof(int));
    if(rc <= 0){
      perror("Read");
      return(rc);
    }
    if(rc != sizeof(int)){
      return(-1);
    }
    *type = ntohl(*type);

    if(length > 0){
      rc = read(fd, message, length);
      if(rc <= 0){
        perror("Read");
        return(rc);
      }
      if(rc != length){
        for(read_length = rc; read_length != length; read_length += rc){
          rc = read(fd, message+read_length, length-read_length);
          if(rc <= 0){
            perror("Read");
            return(rc);
          }
        }
      }
    }
    message[length] = '\0';

    #ifdef DEBUG_VERBOSE
      fflush(stdout);
      fprintf(stdout, "Recieving: %d|%d: %s\n", length, *type, message);
    #endif
  }
  else{
    *type = PERF_ACK;
  }

  return(1);
}

/*------------------------------------------------------------------------
Function name:	write_message
Input parameters:	int fd, int type, char *message
Return value:	-1 if error, 1 if successful
Description:	Sends a meessage in the proper format on fd.
------------------------------------------------------------------------*/

int write_message(int fd, int type, char *message){

  int rc, temp;
  int length, write_length;

  if(message)
    length = strlen(message)*sizeof(char);
  else
    length = 0;

  if( !( (GUI_fd==fd) && !GUI) ){
    temp = htonl(length);
    rc = write(fd, (void*)&temp, sizeof(int));
    if(rc != sizeof(int)){
      perror("Write");
      return(-1);
    }
    temp = htonl(type);
    rc = write(fd, (void*)&temp, sizeof(int));
    if(rc != sizeof(int)){
      perror("Write");
      return(-1);
    }
    if(length != 0){
      rc = write(fd, message, length);
      if(rc < 0){
        perror("Write");
        return(-1);
      }
      if(rc != length){
        for(write_length = rc; write_length != length; write_length += rc){
          rc = write(fd, message+write_length, length-write_length);
          if(rc < 0){
            perror("Write");
            return(-1);
          }
        }
      }
    }
  }

  if(output_file){
    fflush(output_file);
    if(length)
      fprintf(output_file, "%d %d %s\n", length, type, message);
    else
      fprintf(output_file, "%d %d\n", length, type);
  }

  #ifdef DEBUG_VERBOSE
    fflush(stdout);
    if(length)
      fprintf(stdout, "Sending: %d|%d: %s\n", length, type, message);
    else
      fprintf(stdout, "Sending: %d|%d\n", length, type);
  #endif

  return(1);
}

/*------------------------------------------------------------------------
Function name:	gui_to_comm
Input parameters:	None
Return value:	None
Description:	Passes data to GUI.
------------------------------------------------------------------------*/

void* gui_to_comm(void *temp){

  int rc;
  struct App *temp_app;
  struct sigaction read_action;
  const static int readsig = SIGIO;
  
  #ifdef DEBUG
    fprintf(stdout, "Gui_to_comm thread started.\n");
  #endif

  if(GUI){
    if(fcntl(data->gui->fd_from, F_SETOWN, getpid()) == -1){
      perror("Fcntl");
      exit(-1);
    }
    if(fcntl(data->gui->fd_from, F_SETFL, FASYNC) == -1){
      perror("Fcntl");
      exit(-1);
    }
    if(signal(readsig, SIG_IGN) != SIG_DFL){
      fprintf(stderr,"User signal handler detected.\n");
      exit(-1);
    }
    memset(&read_action, 0x00, sizeof(struct sigaction));
    read_action.sa_handler = gui_to_app_comm;
    read_action.sa_flags = SA_RESTART;
    if(sigaction(readsig, &read_action, NULL) < 0){
      perror("Sigaction");
      exit(-1);
    }
  }

  while(1){

    /* Send and delete a record (if any) for all Applications */
    for_apps(temp_app){
      if(temp_app->num_records > 0 && temp_app->rec_list != NULL){

        pthread_mutex_lock(&data->rec_mutex);
          rc = send_record(data->gui, temp_app);
        pthread_mutex_unlock(&data->rec_mutex);

        if(rc != 1){
          fprintf(stderr, "Send_record failed in gui_to.\n");
        }

        pthread_mutex_lock(&data->rec_mutex);
          rc = delete_record(temp_app);
        pthread_mutex_unlock(&data->rec_mutex);

        if(rc != 1){
          fprintf(stderr, "Delete_record failed.\n");
        }
      }
    }

    /* Sleep for 1 microsecond to facilitate a thread switch */
    usleep(1);
  }
}

/*------------------------------------------------------------------------
Function name:  gui_to_app_comm
Input parameters:       None
Return value:	None
Description:	Reads data from GUIand sends it to applications.
------------------------------------------------------------------------*/

static void gui_to_app_comm(){

  int rc;
  int length, size;
  char *mp;
  char list[128];
  char rank[16];
  struct App *temp_app;

  /* Read a message from GUI */
  rc = create_record(data->gui);
  if(rc == 0){
    fprintf(stderr, "Create_record failed.\n");
    fprintf(stderr, "GUI exited.\n");
    fprintf(stderr, "Server exiting.\n");
    exit(1);
  }
  else if(rc != 1){
    fprintf(stderr, "Create_record failed.\n");
  }

  /* Parse message to find Applications to send to */
  mp = data->gui->rec_list->message+1;
  length = strcspn(mp, PERF_DELIM);
  strncpy(list, mp, length);
  list[length] = '\0';

  /* If PERF_ANY, send message to all Applications */
  if(strcmp(list, "PERF_ANY") == 0){
    for_apps(temp_app){

      pthread_mutex_lock(&data->rec_mutex);
        rc = send_record(temp_app, data->gui);
      pthread_mutex_unlock(&data->rec_mutex);

      if(rc != 1){
        fprintf(stderr, "Send_record failed.\n");
      }
    }
  }

  /* Parse each Application number and send message */
  else{
    size = strlen(list);
    for(mp=list; mp<=(list+size); mp+=(length+1)){
      length = strcspn(mp, RANK_DELIM);
      strncpy(rank, mp, length);
      rank[length] = '\0';

      for_apps(temp_app){
        if(atoi(rank) == temp_app->rank){

          pthread_mutex_lock(&data->rec_mutex);
            rc = send_record(temp_app, data->gui);
          pthread_mutex_unlock(&data->rec_mutex);

          if(rc != 1){
            fprintf(stderr, "Send_record failed.\n");
          }
          break;
        }
      }
    }
  }

  pthread_mutex_lock(&data->rec_mutex);
    rc = delete_record(data->gui);
  pthread_mutex_unlock(&data->rec_mutex);

  if(rc != 1){
    fprintf(stderr, "Delete_record failed.\n");
  }
}

/*------------------------------------------------------------------------
Function name:	app_from_comm
Input parameters:	None
Return value:	None
Description:	Receives data from an application.
------------------------------------------------------------------------*/

void* app_from_comm(){

  int rc;
  char message[128];
  fd_set readfds;
  struct timeval tv;
  struct App *temp_app;

  #ifdef DEBUG
    fprintf(stdout, "App_from_comm thread started.\n");
  #endif

  while(1){

    /* Copy masterfds to readfds for temp use */
    memcpy(&readfds, &data->masterfds, sizeof(fd_set));
    tv.tv_sec = SELECT_TIME_LIMIT;
    tv.tv_usec = 0;

    /* Call select to scan all Apps for something sent */
    rc = select(FD_SETSIZE, &readfds, NULL, NULL, &tv);
    if(rc == 0){
      fprintf(stderr, "Select limit of %d seconds expired.\n",
        SELECT_TIME_LIMIT);
    }
    else if(rc < 0){
      perror("Select");
      continue;
    }

    /* Loop for all Applications */
    for_apps(temp_app){

      /* Read from App if something is waiting to be read */
      if(FD_ISSET(temp_app->fd_from, &readfds)){
        if(temp_app->rec_limit_reached == 0 &&
          temp_app->num_records >= REC_BUF_LIMIT){
          temp_app->rec_limit_reached = 1;
          fprintf(stdout, "Buffer limit reached for rank %d.\n",
            temp_app->rank);
          continue;
        }
        else if(temp_app->rec_limit_reached == 1 &&
          temp_app->num_records > REC_BUF_RESTART){
          /*
          fprintf(stdout, "Buffer clearing for rank %d.\n", temp_app->rank);
          */
          continue;
        }
        else if(temp_app->rec_limit_reached == 1 &&
          temp_app->num_records <= REC_BUF_RESTART){
          temp_app->rec_limit_reached = 0;
          fprintf(stdout, "Buffer restarted for rank %d.\n", temp_app->rank);
        }

        rc = create_record(temp_app);
        if(rc == 0){
          fprintf(stderr, "Create_record failed.\n");
          fprintf(stderr, "Application disconnected or terminated.\n");

          /*
          sprintf(message, "%s%d", PERF_DELIM, temp_app->rank);
          rc = write_message(data->gui->fd_to, PERF_DIED, message);
          */

          sprintf(message, "%s%d%s0%s0%s0", PERF_DELIM, temp_app->rank,
            PERF_DELIM, PERF_DELIM, PERF_DELIM);
          rc = write_message(data->gui->fd_to, PERF_BREACH, message);

          if(rc != 1)
            fprintf(stderr, "Write_message failed.\n");

          pthread_mutex_lock(&data->rec_mutex);
            rc = delete_app(temp_app);
          pthread_mutex_unlock(&data->rec_mutex);

          if(rc != 1)
            fprintf(stderr, "Delete_app failed.\n");
        }
        else if(rc < 0){
          fprintf(stderr, "Create_record failed.\n");
        }
      }
    }
  }
}

/*------------------------------------------------------------------------
Function name:	create_record
Input parameters:	struct App *app
Return value:	0 or -1 on error, 1 if successful
Description:	Creates storage for new record from socket fd.
------------------------------------------------------------------------*/

int create_record(struct App *app){

  int rc;
  int type;
  char message[MAX_BUFFER_SIZE];
  struct Record *new_record;

  /* Read from fd into message */
  rc = read_message(app->fd_from, &type, message);
  if(rc <= 0){
    fprintf(stderr, "Read_message failed.\n");
    return(rc);
  }

  /* Create new record and initialize values */
  new_record = (struct Record*)malloc(sizeof(struct Record));
  if(new_record == NULL){
    perror("Malloc");
    return(-1);
  }
  new_record->message = strdup(message);
  new_record->type = type;
  new_record->next = NULL;

  /* Store record */
  pthread_mutex_lock(&data->rec_mutex);

    if(app->num_records == 0 && app->rec_list == NULL)
      app->rec_list = new_record;
    else
      app->rec_list_end->next = new_record;

    app->rec_list_end = new_record;
    app->num_records++;

  pthread_mutex_unlock(&data->rec_mutex);

  return(1);
}

/*------------------------------------------------------------------------
Function name:	send_record
Input parameters:	struct App *rank_to, struct App *rank_from
Return value:	-1 if error, 1 if successful
Description:	Sends record on socket fd.
------------------------------------------------------------------------*/

int send_record(struct App *rank_to, struct App *rank_from){

  int i, rc;
  int size, length;
  float values[32];
  char *mp;
  char message[MAX_BUFFER_SIZE];
  char port_num[64];
  struct Record *record;
  struct Threshhold *temp_thhd;
  struct App *new_app;

  if(rank_to != rank_from){
    record = rank_from->rec_list;
  }
  else{
    fprintf(stderr, "Error: Ranks not set properly in send_record.\n");
    return(-1);
  }

  /* Break down by type */
  switch(record->type){

    case PERF_BREACH:
      break;
    case PERF_CHANGE_EVENT:
      break;
    case PERF_DATASTREAM:
      break;
    case PERF_DIED:
      break;
    case PERF_ERROR:
      break;
    case PERF_MPI:
      break;
    case PERF_OK:
      break;
    case PERF_PAUSE:
      break;
    case PERF_POLL:
      break;

    case PERF_ABORT:
      rc = write_message(rank_to->fd_to, record->type, record->message);
      if(rc != 1){
        fprintf(stderr, "Write_message failed.\n");
        rc = delete_app(rank_to);
          if(rc != 1){
          fprintf(stderr, "Delete_app failed.\n");
        }
        return(-1);
      }
      rc = delete_app(rank_to);
      if(rc != 1){
        fprintf(stderr, "Delete_app failed.\n");
        return(-1);
      }
      return(-1);

    case PERF_END:
      rc = delete_app(rank_from);
      if(rc != 1){
        fprintf(stderr, "Delete_app failed.\n");
      }
      break;

    case PERF_ADDAPP:
      i = 1;
      size = strlen(record->message);
      for(mp=(record->message+1); mp<=(record->message+size);
        mp+=(length+1), i++){
        length = strcspn(mp, PERF_DELIM);
        if(i%2 == 1){
          strncpy(message, mp, length);
          message[length] = '\0';
        }
        if(i%2 == 0){
          strncpy(port_num, mp, length);
          port_num[length] = '\0';
        }
        new_app = init_app(message, port_num);
        if(new_app == NULL){
          fprintf(stderr, "Init_app failed.\n");
          sprintf(message, "%s%s%s%d", RANK_DELIM, new_app->hostname,
            RANK_DELIM, new_app->port_from);
          rc = write_message(data->gui->fd_to, PERF_ERROR, message);
          if(rc != 1){
            fprintf(stderr, "Write_message failed.\n");
          }
          return(-1);
        }
        rc = auth_app(new_app);
        if(rc != 1){
          fprintf(stderr, "Auth_app failed.\n");
          sprintf(message, "%s%s%s%d", RANK_DELIM, new_app->hostname,
            RANK_DELIM, new_app->port_from);
          rc = write_message(data->gui->fd_to, PERF_ERROR, message);
          if(rc != 1){
            fprintf(stderr, "Write_message failed.\n");
          }
          free(new_app);
          return(-1);
        }
        rc = add_app(new_app);
        if(rc != 1){
          fprintf(stderr, "Add_app failed.\n");
          sprintf(message, "%s%s%s%d", RANK_DELIM, new_app->hostname,
            RANK_DELIM, new_app->port_from);
          rc = write_message(data->gui->fd_to, PERF_ERROR, message);
          if(rc != 1){
            fprintf(stderr, "Write_message failed.\n");
          }
          return(-1);
        }
        sprintf(message, "%s%s%s%d%s%d", RANK_DELIM, new_app->hostname,
          RANK_DELIM, new_app->port_from, RANK_DELIM, new_app->rank);
        rc = write_message(data->gui->fd_to, PERF_OK, message);
        if(rc != 1){
          fprintf(stderr, "Write_message failed.\n");
          return(-1);
        }
      }
      return(1);

    case PERF_DATA:
      if(rank_from->thhd_list == NULL || rank_from == data->gui)
        break;
      i = 1;
      size = strlen(record->message);
      for(mp=(record->message+1); mp!='\0' && mp<=(record->message+size);
        mp+=(length+1), i++){
        length = strcspn(mp, PERF_DELIM);
        strncpy(message, mp, length);
        message[length] = '\0';
        if(i >= 6 && (i-6) == rank_from->thhd_list->data_num){
          sscanf(message, "%f", &values[i-6]);
        }
      }
      temp_thhd = rank_from->thhd_list;
      while(temp_thhd != NULL){
        if(temp_thhd->breach == 1){
          temp_thhd = temp_thhd->next;
          continue;
        }
        if(values[temp_thhd->data_num] < temp_thhd->min)
          temp_thhd->var_count++;
        else if(values[temp_thhd->data_num] > temp_thhd->max)
          temp_thhd->var_count++;
        if(temp_thhd->var_count == temp_thhd->variance){
          temp_thhd->breach = 1;
          rc = write_message(rank_from->fd_to, PERF_PAUSE, "");
          if(rc != 1){
            fprintf(stderr,"Write_message failed.\n");
            return(-1);
          }
          sprintf(message, "%s%d%s%f%s%f%s%f", RANK_DELIM, rank_from->rank,
            RANK_DELIM, temp_thhd->min, RANK_DELIM, temp_thhd->max,
            RANK_DELIM, values[temp_thhd->data_num]);
          rc = write_message(data->gui->fd_to, PERF_BREACH, message);
          if(rc != 1){
            fprintf(stderr, "Write_message failed.\n");
            return(-1);
          }
        }
        temp_thhd = temp_thhd->next;
      }
      return(1);

    case PERF_POPUP:
      if(rank_to->thhd_list != NULL){
        free(rank_to->thhd_list);
      }
      rank_to->thhd_list = (struct Threshhold*)malloc
        (sizeof(struct Threshhold));
      if(rank_to->thhd_list == NULL){
        perror("Malloc");
        return(-1);
      }
      rank_to->thhd_list->breach = 0;
      rank_to->thhd_list->var_count = 0;
      rank_to->thhd_list->prev = NULL;
      rank_to->thhd_list->next = NULL;
      i = 0;
      size = strlen(record->message);
      for(mp=(record->message+1); i<=4 && mp<=(record->message+size);
        mp+=(length+1), i++){
        length = strcspn(mp, PERF_DELIM);
        strncpy(message, mp, length);
        message[length] = '\0';
        if(i == 1)
          sscanf(message, "%f", &rank_to->thhd_list->min);
        else if(i == 2)
          sscanf(message, "%f", &rank_to->thhd_list->max);
        else if(i == 3)
          rank_to->thhd_list->variance = atoi(message);
        else if(i == 4)
          rank_to->thhd_list->data_num = atoi(message);
      }
      return(1);

    case PERF_STARTAPP:
      if(rank_to->thhd_list == NULL){
        break;
      }
      if(rank_to->thhd_list->breach == 0){
        break;
      }
      else{
        free(rank_to->thhd_list);
        rank_to->thhd_list = NULL;
        break;
      }

    default:
      fprintf(stderr, "Message type %d not recognized.\n", record->type);
      return(-1);
  }

  rc = write_message(rank_to->fd_to, record->type, record->message);
  if(rc != 1){
    fprintf(stderr, "Write_message failed.\n");
    return(-1);
  }

  return(1);
}

/*------------------------------------------------------------------------
Function name:	delete_record
Input parameters:	struct App *app
Return value:	-1 if error, 1 if successful
Description:	Removes and frees record from app_list[rank].
------------------------------------------------------------------------*/

int delete_record(struct App *app){

  struct Record *record;

  record = app->rec_list;
  if(record == NULL){
    return(-1);
  }
  app->rec_list = record->next;
  app->num_records--;
  if(app->num_records == 0){
    app->rec_list_end = NULL;
  }

  free(record->message);
  free(record);
  record = NULL;

  return(1);
}

/*------------------------------------------------------------------------
Function name:	add_app
Input parameters:	struct App *app
Return value:	-1 if error, 1 if successful
Description:	Adds an app to the app_list.
------------------------------------------------------------------------*/

int add_app(struct App *app){

  data->app_list->prev->next = app;
  app->prev = data->app_list->prev;
  data->app_list->prev = app;
  app->next = data->app_list;

  data->num_apps++;

  /*
  Set fd in list for reading from 
  FD_SET(app->fd_from, &data->masterfds);
  */

  return(1);
}

/*------------------------------------------------------------------------
Function name:	delete_app
Input parameters:	struct App *app
Return value:	-1 if error, 1 if successful
Description:	Removes and frees app from app_list.
------------------------------------------------------------------------*/

int delete_app(struct App *app){

  if(app == data->app_list){
    return(1);
  }

  FD_CLR(app->fd_from, &data->masterfds);

  while(delete_record(app) != -1)
    fprintf(stdout, "Deleting record.\n");
  if(app->thhd_list != NULL)
    free(app->thhd_list);

  app->prev->next = app->next;
  app->next->prev = app->prev;

  data->num_apps--;
  if(data->num_apps <= 0){
    fprintf(stderr, "All applications exited.\n");
    exit(1);
  }

  close(app->fd_to);
  close(app->fd_from);

  return(1);
}
