cc file  update.f  contains:   proflux.f  flux.f  pde.f

c########1###########################################################7##
c with void
c for Fs-Px 

       SUBROUTINE PROFLUX( CL,CS,alam,T, sratioW, Ip,Iw,Jp,Jw,Jv,Dr,Dz, 
     &                                   akWmax,   Reff, h, cp, sratio )
c______________________________________________________________lines  96
c12-23-93 changed formula for akW(T)= akW0 + akW1 * [T-TE]
c 6-19-93 variable akS(T),  akL(T) external
c 8-17-92,variable akW(T)= akW0 + akW1 * [T-Tbry]
c 6-17-92  modified for  Diopside - Anorthite
c10- 4-90    added void (free) space, j=0,-1,...,-Jv+1
c 4-21-90  new fitting of hL
c 4- 7-90  new fittings of cpL, hL 
c12-15-89   find Reff(i,j) and h(i,j) for FLUX(), and sratio for dt
c---------
c inputs:  CL,CS,alam,T,akWmax  from PHASE();    external hL, HbS(x,T)
c          sratioW from MESH(); rho from common 
c outputs:  Reff,h  needed in FLUX()
c           sratio  needed in MAIN to set the time-step dt.
c********
                implicit double precision (a-h,o-z)
      EXTERNAL  cpL, cpS, hL, hS, akL, akS

      DIMENSION CL(Ip,Jp), CS(Ip,Jp), alam(Ip,Jp) 
      DIMENSION Reff(Ip,Jp), h(Ip,Jp), cp(Ip,Jp) 
      DIMENSION T(Ip+Iw+1,1-Jw-Jv:Jp+Jw+1)

      common/weights/ WA,WB,Wdiff, TA,TB, xE,TE,Tref,DhfusE,xinit
      common/propts1/ rho,cokL(3),cokS(2),rhoW,cpW,akW0,akW1,akV
      common/propts2/ DL,DS,delta,beta
      common/ coefs / coCpS(7), coCpL(3), cohL(6), cohhLig(6)
      common/switch / T32kL

cc.........to translate C=weight fraction to x=mole fraction B..........
           XX( Cij ) = WA * Cij / ( WB - Wdiff * Cij )

cc.....................to find the formula weight.......................
cd           WM( xij ) = WA + Wdiff * xij

c*************** find Reff(i,j),h(i,j), for all PCM nodes **************
#if defined(PROFOMETER)
       call mark_profometer('MAGENTA\0')
#elif defined(PERFOMETER)
#include "perfcolorcodes.h"
       call mark_perfometer(MAGENTA, "update.F")
#endif
       Do 100 j = 1 , Jp
         Do 150 i = 1 , Ip

c        -------------------------- find  Reff--------------------------
              Tij = T(i,j)
               TK = Tij + 273.15 
             akLij = akL( Tij )
             akSij = akS( Tij )
c>>should use Kirchoff instead of this:
cc  by adding resistivities (harmonic average of conductivities):
c          Reff(i,j) = alam(i,j) / akLij + ( 1.d0-alam(i,j) ) / akSij
c>> by averaging conductivities:
          Reff(i,j) = 1. / (alam(i,j)*akLij + ( 1.d0-alam(i,j) )*akSij)

             CLij = CL(i,j)
             CSij = CS(i,j)
c        ----------------------find xLij from CL(i,j)-------------------
             xLij = XX( CLij ) 
cd            WxLij = WM( xLij )
c        ----------------------find xSij from CS(i,j)-------------------
             xSij = XX( CSij )
cd            WxSij = WM( xSij ) 

c        --------------also find cp(i,j), needed to find dt later-------
cc.      ..............from EXTERNAL functions cpL(), cpS().............
            cpLxLT = cpL( xLij, Tij )
            cpSxST = cpS( xSij, Tij )

          cp(i,j) = alam(i,j) * cpLxLT + (1.d0-alam(i,j)) * cpSxST

c        ---------------------------to find h(i,j)--------------------
c           ........................ find hL(xL,T)....EXTERNAL hL
             hLxLT = hL( xLij, Tij )
c            ....................... find hS(xS,T)....EXTERNAL HbS
             hSxST = hS( xSij, Tij ) 
c            ...........................and finally ...........
          h(i,j) = alam(i,j) * hLxLT + (1.d0-alam(i,j)) * hSxST

  150     CONTINUE
100    CONTINUE
             
cc------------------- find stable step-size --------------------------
cc            sratioW from MESH(), akWmax from PHASE()
                 sratW = sratioW / akWmax
                sratio = sratW 
       DO 200 j = 2, Jp-1
          DO 250 i = 2, Ip-1
                 Rmini = MIN( Reff(i-1,j), Reff(i,j), Reff(i+1,j) )
                 Rminj = MIN( Reff(i,j-1), Reff(i,j), Reff(i,j+1) )
                 AVRij = 2*( 1/(Dr*Dr)/Rmini + 1/(Dz*Dz)/Rminj ) 
                srat = rho * cp(i,j) / AVRij
                  sratio = MIN( srat, sratio )
  250     CONTINUE
200    CONTINUE

c... this smallest sratio is used in MAIN to set: dt = CFL * sratio... 
                 return
                  end


c########1#########2#########3#########4#########5#########6#########7##
c with void  ( non-conducting )
c variable akW(T)= k0 + k1*(T-TE))

       SUBROUTINE FLUX( C,CL,CS,T,alam,h,Reff,Ar,Az,r,r2,z,z2,Tbry,TE,
     &                 Ip,Iw,Jp,Jw,Jv, Dr,DrW,Dz,DzW,DzV, drokIp,dzokJv,
     &                                           ArJ,AzJ,ArQ,AzQ,QAout )
c______________________________________________________________lines 297
c 8-17-92,12-23-93  variable akW(T)= akW0 + akW1*[T-TE]
c11- 9-90  changed void to have uniform Tvoid (rhoV=0),non-conducting
c10- 6-90  added void (free) space, j=0,-1,...,-Jv+1, conducting
c 9-28-90  corrected rminL,rminS,zminL,zminS: Clam instead of alam
c 9-25-90  corrected ()/DrW/2,()/DzW/2  to ()/DrW2,()/DzW2
c                    (discovered by James Smith !)
c 4-29-90  added AzQ in lateral wall !!! it was missing !!!
c 9-23-90  added AzQ in void
c 12-15-89      computes all fluxes needed in PDE(): ArJ,AzJ,ArQ,AzQ
c               assumes imposed Tbry on outside surfaces of container.
c inputs:C                 from PDE() at previous time-step,
c        CL,CS,T,alam,     from PHASE() at previous time-step,
c        h,Reff            from PROFLUX(),
c        Ar,Az,r,r2,z,z2,drokIp,dzokJv from MESH() 
c        Dr,DrW, Dz,DzW,DzV            from MESH() 
c        rho,DL,DS,akW0,akW1,akV,delta,beta via common/propts/ 
c        Tbry         from MAIN() coming from INPUT().
c outputs: ArJ,AzJ,ArQ,AzQ , needed in PDE().
c          QAout          needed in OUTPUT() for heat balance 
c--------
                implicit double precision (a-h,o-z)
      DIMENSION  r(Ip+Iw),    z(1-Jw-Jv:Jp+Jw)
      DIMENSION r2(Ip+Iw+1), z2(1-Jw-Jv:Jp+Jw+1)
      DIMENSION Ar(Ip+Iw+1,1-Jw-Jv:Jp+Jw)
      DIMENSION Az(Ip+Iw,1-Jw-Jv:Jp+Jw+1)
      DIMENSION  C(Ip,Jp), CL(Ip,Jp), CS(Ip,Jp), alam(Ip,Jp) 
      DIMENSION   T(Ip+Iw+1,1-Jw-Jv:Jp+Jw+1)
      DIMENSION ArJ(Ip+Iw+1,1-Jw-Jv:Jp+Jw)
      DIMENSION AzJ(Ip+Iw,1-Jw-Jv:Jp+Jw+1)
      DIMENSION ArQ(Ip+Iw+1,1-Jw-Jv:Jp+Jw)
      DIMENSION AzQ(Ip+Iw,1-Jw-Jv:Jp+Jw+1)
      DIMENSION Reff(Ip,Jp), h(Ip,Jp)

      common/propts1/ rho,cokL(3),cokS(2),rhoW,cpW,akW0,akW1,akV
      common/propts2/ DL,DS,delta,beta
      common/switch / T32kL


cc---------------------- formula for variable akW ---------------------
c version 1: akW( Tij ) = akW0 * ( 1. + akW1 * ( Tij - Tbry ) )
c version 2:
             akW( Tij ) = akW0  + akW1 * ( Tij - TE )

             DrW2 = DrW/2
             DzW2 = DzW/2
c------------------------ SET INTERNAL FLUXES -------------------------
c......................ArJ, ArQ for j=1 (bottom).........
#if defined(PROFOMETER)
       call mark_profometer('BLUE\0')
#elif defined(PERFOMETER)
       call mark_perfometer(BLUE, "update.F")
#endif
       Do 101 i =2,Ip
c>old>:
c          rminL = min(alam(i-1,1),alam(i,1))
c          rminS = min( 1-alam(i-1,1) , 1-alam(i,1) )
c>new>:
            Clam1 = ( CS(i-1,1)-C(i-1,1) ) / ( CS(i-1,1)-CL(i-1,1) )
            Clam2 = ( CS(i,1)-C(i,1) ) / ( CS(i,1)-CL(i,1) )
            rminL = min( Clam1 , Clam2 )
            rminS = min( 1.0-Clam1 , 1.0-Clam2 )
	  rJ1=rminL*DL*(CL(i-1,1)-CL(i,1))+rminS*DS*(CS(i-1,1)-CS(i,1))
	       rJ = ( rJ1 + delta * ( T(i,1)- T(i-1,1) ) ) / Dr
         ArJ(i,1) = Ar(i,1) * rJ
               rR = (r2(i)-r(i-1))*Reff(i-1,1) + (r(i)-r2(i))*Reff(i,1)
               rG = ( T(i-1,1) - T(i,1) ) / rR
              rQG = rG + beta*(C(i,1) -C(i-1,1))/Dr 
              rQJ = rJ * rho * .5d0*( h(i-1,1) + h(i,1) ) 
	 ArQ(i,1) = Ar(i,1) * ( rQG + rQJ )
101    CONTINUE
c......................AzJ, AzQ for i=1 (left)............
       Do 121 j=2,Jp
c>old>:
c          zminL = min(alam(1,j-1),alam(1,j))
c          zminS = min(1-alam(1,j-1),1-alam(1,j))
c>new>:
            Clam1 = ( CS(1,j-1)-C(1,j-1) ) / ( CS(1,j-1)-CL(1,j-1) )
            Clam2 = ( CS(1,j)-C(1,j) ) / ( CS(1,j)-CL(1,j) )
            zminL = min( Clam1 , Clam2 )
            zminS = min( 1.0-Clam1 , 1.0-Clam2 )
          zJ1=zminL*DL*(CL(1,j-1)-CL(1,j))+zminS*DS*(CS(1,j-1)-CS(1,j))
  	       zJ = ( zJ1 + delta * ( T(1,j) - T(1,j-1) ) ) / Dz
	 AzJ(1,j) = Az(1,j) * zJ
	       zR = (z2(j)-z(j-1))*Reff(1,j-1)+(z(j)-z2(j))*Reff(1,j)
               zG = ( T(1,j-1) - T(1,j) ) / zR
	      zQG = zG + beta*(C(1,j) -C(1,j-1))/Dz 
	      zQJ = zJ * rho * .5d0*( h(1,j-1) + h(1,j) ) 
         AzQ(1,j) = Az(1,j) * ( zQG + zQJ )
121    Continue

c......................ArJ,AzJ, ArQ,AzQ for all interior nodes........
       Do 150 i =2,Ip
         Do 160 j=2,Jp

c>old>:
c          rminL = min(alam(i-1,j),alam(i,j))
c          rminS = min(1 - alam(i-1,j),1 - alam(i,j))
c>new>:
            Clam1 = ( CS(i-1,j)-C(i-1,j) ) / ( CS(i-1,j)-CL(i-1,j) )
            Clam2 = ( CS(i,j)-C(i,j) ) / ( CS(i,j)-CL(i,j) )
            rminL = min( Clam1 , Clam2 )
            rminS = min( 1.0-Clam1 , 1.0-Clam2 )
          rJ1=rminL*DL*(CL(i-1,j)-CL(i,j))+rminS*DS*(CS(i-1,j)-CS(i,j))
	       rJ = ( rJ1 + delta * ( T(i,j)- T(i-1,j) ) ) / Dr
	 ArJ(i,j) = Ar(i,j) * rJ
c>old>:
c          zminL = min(alam(i,j-1),alam(i,j))
c          zminS = min(1-alam(i,j-1),1-alam(i,j))
c>new>:
            Clam1 = ( CS(i,j-1)-C(i,j-1) ) / ( CS(i,j-1)-CL(i,j-1) )
            Clam2 = ( CS(i,j)-C(i,j) ) / ( CS(i,j)-CL(i,j) )
            zminL = min( Clam1 , Clam2 )
            zminS = min( 1.0-Clam1 , 1.0-Clam2 )
          zJ1=zminL*DL*(CL(i,j-1)-CL(i,j))+zminS*DS*(CS(i,j-1)-CS(i,j))
               zJ = ( zJ1 + delta * ( T(i,j) - T(i,j-1) ) ) / Dz
         AzJ(i,j) = Az(i,j) * zJ

c      ...from PROFLUX : Reff(i,j) = lamda(i,j)/kL(i,j)
c      ............................     + (1-lamda(i,j))/kS(i,j).......

	       rR = (r2(i)-r(i-1))*Reff(i-1,j) + (r(i)-r2(i))*Reff(i,j)
               zR = (z2(j)-z(j-1))*Reff(i,j-1) + (z(j)-z2(j))*Reff(i,j)
               rG = ( T(i-1,j) - T(i,j) ) / rR
	       zG = ( T(i,j-1) - T(i,j) ) / zR
	      rQG = rG + beta *(C(i,j) -C(i-1,j))/Dr 
	      rQJ = rJ * rho * .5d0 * (h(i-1,j) + h(i,j)) 
         ArQ(i,j) = Ar(i,j) * ( rQG + rQJ )
              zQG = zG + beta *(C(i,j) -C(i,j-1))/Dz 
	      zQJ = zJ * rho * .5d0 * (h(i,j-1) + h(i,j)) 
	 AzQ(i,j) = Az(i,j) * ( zQG + zQJ )

 160     CONTINUE
150    CONTINUE

c---------------SET fluxes at interior wall surfaces-------------------

c......................ArJ, ArQ for i=1(axis) and i=Ip+1.............
       Do 200 j = 1,Jp
             ArJ(1,j)    = 0.d0
             ArJ(Ip+1,j) = 0.d0
             ArQ(1,j)    = 0.d0
                     Tij = T(Ip+1,j)
                   akWij = akW( Tij )
              rR = (r2(Ip+1)-r(Ip))*Reff(Ip,j)+(r(Ip+1)-r2(Ip+1))/akWij 
             ArQ(Ip+1,j) = Ar(Ip+1,j) * ( T(Ip,j)-T(Ip+1,j) ) / rR
     &            
200    Continue
c......................AzJ, AzQ for j=1(bottom) and j=Jp+1(top).....
       Do 300 i= 1, Ip
             AzJ(i,1)    = 0.d0
             AzJ(i,Jp+1) = 0.d0
              zR1 = ( z2(1)-z(0) )/akV + ( z(1)-z2(1) ) * Reff(i,1)
             AzQ(i,1)    = Az(i,1) * ( T(i,0)-T(i,1) ) / zR1
                     Tij = T(i,Jp+1)
                   akWij = akW( Tij )
              zRIp=(z2(Jp+1)-z(Jp))*Reff(i,Jp)+(z(Jp+1)-z2(Jp+1))/akWij
             AzQ(i,Jp+1) = Az(i,Jp+1) * ( T(i,Jp)-T(i,Jp+1) ) / zRIp
300    CONTINUE                
c---------------set heat fluxes in the walls and on the outside--------
                  QAout =  0.0
                   Jv1  = -Jv+1
c............................ ArQ in bottom and top wall ..............
      Do 500 j = 1, Jw 
		
c.       ...................at the axis (i=1): set 0.0 there:
            ArQ(1,Jv1-j) = 0.d0
            ArQ(1,Jp+j)  = 0.d0 
c.       ...................ArQ inside bottom and top walls........
         Do 550 i = 2, Ip+Iw
                    Tim1 = T(i-1,Jv1-j)
                  akWim1 = akW( Tim1 )
                    Ti   = T( i,Jv1-j)
                  akWi   = akW( Ti )
              rR = (r2(i)-r(i-1))/akWim1 + (r(i)-r2(i-1))/akWi
            ArQ(i,Jv1-j)= Ar(i,Jv1-j) * (T(i-1,Jv1-j)-T(i,Jv1-j)) / rR
                    Tim1 = T(i-1,Jp+j)
                  akWim1 = akW( Tim1 )
                    Ti   = T( i,Jp+j)
                  akWi   = akW( Ti )
              rR = (r2(i)-r(i-1))/akWim1 + (r(i)-r2(i-1))/akWi
            ArQ(i,Jp+j)=Ar(i,Jp+j) * ( T(i-1,Jp+j)-T(i,Jp+j) ) / rR
  550    CONTINUE
c.       ...................ArQ at the outside: set Tbry there
                    Ti   = T(Ip+Iw,Jv1-j)
                  akWi   = akW( Ti )
          ArQ(Ip+Iw+1,Jv1-j)= Ar(Ip+Iw+1,Jv1-j)*akWi
     &                                     *(T(Ip+Iw,Jv1-j)-Tbry)/DrW2
                  QAout = QAout + ArQ(Ip+Iw+1,Jv1-j)
                    Ti   = T(Ip+Iw,Jp+j)
                  akWi   = akW( Ti )
          ArQ(Ip+Iw+1,Jp+j)= Ar(Ip+Iw+1,Jp+j)*akWi
     &                                    *( T(Ip+Iw,Jp+j)- Tbry )/DrW2
                  QAout = QAout + ArQ(Ip+Iw+1,Jp+j)

500   CONTINUE
c...........................ArQ in lateral wall....................
      DO 600  j = -Jv+1, Jp

c.       ..........................inside middle part........
         DO 650  i = Ip+2, Ip+Iw
                    Tim1 = T(i-1,j)
                  akWim1 = akW( Tim1 )
                    Ti   = T( i,j)
                  akWi   = akW( Ti )
              rR = (r2(i)-r(i-1))/akWim1 + (r(i)-r2(i-1))/akWi
                 ArQ(i,j) = Ar(i,j) * ( T(i-1,j) - T(i,j) ) / rR
  650    CONTINUE
c.       ....................at the outside: set Tbry there.........
                    Ti   = T(Ip+Iw,j)
                  akWi   = akW( Ti )
         ArQ(Ip+Iw+1,j) = Ar(Ip+Iw+1,j)*akWi*( T(Ip+Iw,j)-Tbry )/DrW2
                  QAout = QAout + ArQ(Ip+Iw+1,j)

600   CONTINUE
c...........................AzQ in bottom and top wall...............
      Do 700  i = 1, Ip+Iw

c.       ..................at outside bottom surface: set Tbry
                    Tj   = T( i, Jv1-Jw )
                  akWj   = akW( Tj )
         AzQ(i,Jv1-Jw) = Az(i,Jv1-Jw)*akWj
     &                          *( Tbry-T(i,Jv1-Jw) ) / DzW2
                  QAout = QAout - AzQ(i,Jv1-Jw)
c.       ..................inside bottom wall.....................
         DO 750  j = -Jv-Jw+2, -Jv
                    Tjm1 = T( i,j-1 )
                  akWjm1 = akW( Tjm1 )
                    Tij  = T( i, j )
                  akWij  = akW( Tij )
                  zR = (z2(j)-z(j-1))/akWjm1 + (z(j)-z2(j-1))/akWij
            AzQ(i,j) = Az(i,j) * ( T(i,j-1) - T(i,j) ) / zR
  750    CONTINUE
c.       ..................inside top wall........................
         DO 770  j = Jp+2, Jp+Jw
                    Tjm1 = T( i,j-1 )
                  akWjm1 = akW( Tjm1 )
                    Tij  = T( i, j )
                  akWij  = akW( Tij )
                  zR = (z2(j)-z(j-1))/akWjm1 + (z(j)-z2(j-1))/akWij
            AzQ(i,j) = Az(i,j) * ( T(i,j-1) - T(i,j) ) / zR
  770    CONTINUE
c.       ..................at outside top surface: set Tbry there:
                    Tij  = T( i, Jp+Jw )
                  akWij  = akW( Tij )
            AzQ(i,Jp+Jw+1) = Az(i,Jp+Jw+1)*akWij
     &                                    *( T(i,Jp+Jw)-Tbry ) / DzW2
                   QAout = QAout + AzQ(i,Jp+Jw+1)

700   CONTINUE
c..........................AzQ in lateral wall....................
      DO 800  j = -Jv+1, Jp+1
         DO 850  i = Ip+1, Ip+Iw
                    Tjm1 = T( i,j-1 )
                  akWjm1 = akW( Tjm1 )
                    Tij  = T( i, j )
                  akWij  = akW( Tij )
                  zR = (z2(j)-z(j-1))/akWjm1 + (z(j)-z2(j-1))/akWij
            AzQ(i,j) = Az(i,j) * ( T(i,j-1) - T(i,j) ) / zR
 850     CONTINUE
800   CONTINUE
c-------------------- set radial heat fluxes in void ------------------

      DO 900  j = -Jv+1, 0
c..      ...................for i=1 (axis)...................
            ArQ(1,j) = 0.d0
c..      ...................in interior of void..............
c..       these may be commended out for non-conducting void (equal 0) 
c         DO 920  i = 2, Ip
c            rR = ( r(i) - r(i-1) ) / akV
c            ArQ(i,j) = Ar(i,j) * ( T(i-1,j) - T(i,j) ) / rR
c  920    CONTINUE
c..      .................for i=Ip+1 (right).........................
c..      drokIp = ( r2(Ip+1) - r(Ip) )/akV + ( r(Ip+1)-r2(Ip+1) )/akW
c..      ........          passed thru from mesh()...................
             ArQ(Ip+1,j) = Ar(Ip+1,j)* ( T(Ip,j)-T(Ip+1,j) ) / drokIp
900   CONTINUE
c-------------------- set axial heat fluxes in void ------------------

      DO 950  i = 1, Ip
c..      ...................for j=1 (top)...already done above......
c..      ...................in interior of void.....................
c..       these may be commended out for non-conducting void (equal 0)
c         DO 970  j = 2-Jv, 0
c               zR = ( z(j) - z(j-1) ) / akV
c               AzQ(i,j) = Az(i,j)* ( T(i,j-1) - T(i,j) ) / zR
c  970    CONTINUE
c..      ...............for j = -Jv+1 (bottom)........................
c..      dzokJv = ( z2(1-Jv)-z(-Jv) )/akW + (z(1-Jv)-z2(1-Jv))/akV
c..      .........          passed thru from mesh()................
             AzQ(i,Jv1) = Az(i,Jv1)*( T(i,-Jv)-T(i,Jv1) )/dzokJv
950   CONTINUE
c*****
      return
      end


c########1###########################################################7##
c with void ( non-conducting ) 
       SUBROUTINE PDE( ArJ,AzJ,ArQ,AzQ, V,rho,rhoW,dt, Ip,Iw, Jp,Jw,Jv,
     &                                              C,e, Etotal,Ctotal )
c_______________________________________________________________lines 77
c10-4-90    added void (free) space, j=0,-1,...,-Jv+1
c 6-29-89       updates C(i,j) and e(i,j)
c inputs: ArJ,AzJ,ArQ,AzQ      from FLUX(); 
c         dt ; V,rho,rhoW,rhoV from MAIN() 
c outputs: C(i,j), e(i,j), Etotal, Ctotal
c--------
               implicit double precision (a-h,o-z)
      DIMENSION  C(Ip,Jp), e(Ip+Iw,1-Jw-Jv:Jp+Jw) 
      DIMENSION            V(Ip+Iw,1-Jw-Jv:Jp+Jw)
      DIMENSION ArJ(Ip+Iw+1,1-Jw-Jv:Jp+Jw)
      DIMENSION AzJ(Ip+Iw,1-Jw-Jv:Jp+Jw+1)
      DIMENSION ArQ(Ip+Iw+1,1-Jw-Jv:Jp+Jw)
      DIMENSION AzQ(Ip+Iw,1-Jw-Jv:Jp+Jw+1)
#if defined(PROFOMETER)
      call mark_profometer('red\0')
#elif defined(PERFOMETER)
      call mark_perfometer(RED, "update.F")
#endif
          Etotal = 0.d0
          Ctotal = 0.d0
c---------------------------in the PCM-------------------------------
       DO 100 j = 1, Jp
         DO 150 i= 1, Ip

         C(i,j) = C(i,j)+(ArJ(i,j)-ArJ(i+1,j)+AzJ(i,j)-AzJ(i,j+1))*dt
     &                                                        / V(i,j)
c>>            Ctotal = Ctotal + C(i,j)
            Ctotal = Ctotal + C(i,j) * rho * V(i,j)
          DEij = ( ArQ(i,j)-ArQ(i+1,j)+AzQ(i,j)-AzQ(i,j+1) )*dt
          rhoVij = rho * V(i,j)
         e(i,j) = e(i,j) + DEij / rhoVij 
         Etotal = Etotal + rhoVij * e(i,j)

  150    CONTINUE
100    CONTINUE
c--------------------------in vertical wall--------------------------
       DO 200    j = 1-Jw-Jv, Jp+Jw
         DO 250  i = Ip+1, Ip+Iw

           DEij = ( ArQ(i,j)-ArQ(i+1,j)+AzQ(i,j)-AzQ(i,j+1) )*dt
           rhoWVij = rhoW * V(i,j)
         e(i,j) = e(i,j) + DEij / rhoWVij 
         Etotal = Etotal + rhoWVij * e(i,j)

  250    CONTINUE
200    CONTINUE
c-------------------------in top and bottom walls--------------------
              Jv1 = 1-Jv
       DO 300   j = 1, Jw
         DO 350 i = 1, Ip

       DEij=(ArQ(i,Jp+j)-ArQ(i+1,Jp+j)+AzQ(i,Jp+j)-AzQ(i,Jp+j+1))*dt
            rhoWVij = rhoW * V(i,Jp+j)
         e(i,Jp+j)  = e(i,Jp+j) + DEij / rhoWVij 
             Etotal = Etotal + rhoWVij * e(i,Jp+j)

               DEij = ( ArQ(i,Jv1-j)-ArQ(i+1,Jv1-j)+
     &                  AzQ(i,Jv1-j)-AzQ(i,Jv1-j+1) ) * dt
            rhoWVij = rhoW * V(i,Jv1-j)
         e(i,Jv1-j) = e(i,Jv1-j) + DEij / rhoWVij 
             Etotal = Etotal + rhoWVij * e(i,Jv1-j)

  350    CONTINUE
300    CONTINUE
c-------------------------in the void (if conducting, need rhoV)------
c       DO 400   j = 1-Jv, 0
c         DO 450 i = 1, Ip
c             DEij = (ArQ(i,j)-ArQ(i+1,j)+AzQ(i,j)-AzQ(i,j+1))*dt
c           rhoVVij = rhoV * V(i,j)
c           e(i,j)   = e(i,j) + DEij / rhoVVij
c           Etotal = Etotal + rhoVVij * e(i,j)
c  450    CONTINUE
c400    CONTINUE
                 return
                 end
