/*
 *  This is a simple test to see how much wtimings are off between computers.
 *  First it loops through on each process for several loops to see if it
 *  ever degrades on one node.  Then it starts using the rest of the nodes
 *  to check. 
 *  Kevin London 	April 2, 1998
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <mpi.h>
#include "perfcolorcodes.h"
#include "perfometer.h"
/*
 * This can be set in the makefile as well.
 */

#ifndef NUM_LOOPS
#define NUM_LOOPS 	1000
#endif
#ifndef SMALL_LOOP
#define SMALL_LOOP	50
#endif
#ifndef REVERSE_LOOP
#define REVERSE_LOOP    50
#endif

int
main(argc, argv)
  int argc;
  char *argv[];
{
  MPI_Status status;
  int i, j, rank, size, temp, flag, *mpi_global;
  double t0, t1, t2;

  MPI_Init( &argc, &argv);
  MPI_Comm_rank( MPI_COMM_WORLD, &rank );
  MPI_Comm_size( MPI_COMM_WORLD, &size );
  if ( size > 2 )
  {
	fprintf( stderr, "This tester only needs 1 or 2 processes to run.\n");
  	MPI_Finalize();
	exit(1);
  }
  perfometer();
  if ( rank == 0 )
	mark_perfometer( GREEN, "Rank 0");
  else
	mark_perfometer( YELLOW, "Rank 1");
  /*
   *  Here is the single node test.  Here we don't check to see if the
   *  previous call equals the current one, since the standard doesn't
   *  specify what happens on that.  It simply specifies that the timer
   *  won't go backward and that it will always go forward if MPI_Wtime is
   *  called before a MPI_Send call and after the matching MPI_Recv call
   */
  if ( rank == 0 )
  {
	if ( (t0 = MPI_Wtime( )) == 0 )
        {
#ifdef DEBUG
		fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "SINGLE NODE: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	}
	for ( i = 0; i < NUM_LOOPS; i++ )
        {
	  if ( (t1 = MPI_Wtime( )) == 0 )
	  {
#ifdef DEBUG
		fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "SINGLE NODE: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	  }
	  t2 = t1 - t0;
          if ( t2 < 0 )
	  {
#ifdef DEBUG
		fprintf( stderr, "First Wtime value greater than the next.\n" );
#endif
		fprintf( stderr, "SINGLE NODE: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	  }
	  t0 = t1;
	}
  }
  /*
   * End of single node test and it passed.  We exit out above if it doesn't
   * PASS since there is no reason to check multi if single fails
   */
  if ( rank == 0 )
     fprintf( stderr, "SINGLE NODE: PASSED\n" );
 
  /*
   * Have to check the MPI_WTIME_IS_GLOBAL attribute, if it is set multi-node
   * calls to MPI_Wtime are supposed to be synchronized, if it isn't set then
   * there is no reason to continue with the test.
   */

  MPI_Attr_get( MPI_COMM_WORLD, MPI_WTIME_IS_GLOBAL, &mpi_global, &flag );
  if ( flag == 0 )
  {
	if ( rank == 0 )
	   fprintf( stderr, "No flag associated with MPI_WTIME_IS_GLOBAL\n" );
	MPI_Abort( MPI_COMM_WORLD, -1);
	exit(1);
  }
  if ( !*mpi_global )
  {
	if ( rank == 0 )
	   fprintf( stderr, "Global Sync turned off\n" );
	MPI_Abort( MPI_COMM_WORLD, -1);
	exit(1);
  }
  if ( rank == 0 )
     fprintf( stderr, "Global Sync turned ON proceding with tests\n" );
  fflush( stderr );
  /* Lets let the master process catch up -KSL */
  MPI_Barrier( MPI_COMM_WORLD );
  if ( size < 2 )
  {
	MPI_Abort( MPI_COMM_WORLD, -1);
	exit(1);
  }
  /*  Here we are testing that what the standard says is always true that
   *  If a MPI_Wtime is before a send and after the matching recv it is
   *  always a greater value the second time.
   */
  for ( i = 0; i < SMALL_LOOP; i++ )
  {
	t0=0;t1=0;
        MPI_Barrier( MPI_COMM_WORLD );
	if ( rank == 0 )
	{
	   if ( (t0 = MPI_Wtime()) == 0 )
           {
#ifdef DEBUG
	        fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "MULTI SEND/RECV: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	   }
	   MPI_Send(&t0, 1, MPI_DOUBLE, 1, 0, MPI_COMM_WORLD );
	   MPI_Recv(&t1, 1, MPI_DOUBLE, 1, 0, MPI_COMM_WORLD, &status );
	   t2 = t1 - t0;
#ifdef MORE
	   printf("(%d loop) T0:%f T1:%f  Difference: %f\n", i, t0, t1, t2 );
#endif
           if ( t2 < 0 )
	   {
#ifdef DEBUG
		fprintf( stderr, "First Wtime value greater than the next.\n" );
#endif
		fprintf( stderr, "MULTI SEND/RECV: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	   }
 	   if ( t2 == 0 )
	   {
#ifdef DEBUG
	     fprintf( stderr, "First Wtime value equal to the next.\n" );
#endif
	     fprintf( stderr, "MULTI SEND/RECV: FAILED\n" );
  	     MPI_Abort( MPI_COMM_WORLD, -1);
	     exit(1);
	   }
	}
	else
	{
	   MPI_Recv(&t1, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status );
	   if ( (t0 = MPI_Wtime()) == 0 )
           {
#ifdef DEBUG
		fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "MULTI SEND/RECV: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	   }
	   MPI_Send(&t0, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD );
	}
   }
  if ( rank == 0 )
  	fprintf(stderr, "MULTI SEND/RECV: PASSED\n" );

  /*  Here we are testing that what the standard says is always true that
   *  If a MPI_Wtime is before a send and after the matching recv it is
   *  always a greater value the second time.
   */
  for ( i = 0; i < REVERSE_LOOP; i++ )
  {
	t0=0;t1=0;
        MPI_Barrier( MPI_COMM_WORLD );
	if ( rank == 0 )
	{
           sleep(1);
	   if ( (t0 = MPI_Wtime()) == 0 )
           {
#ifdef DEBUG
	        fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "MULTI SEND/RECV REVERSED: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	   }
	   MPI_Send(&t0, 1, MPI_DOUBLE, 1, 0, MPI_COMM_WORLD );
	   MPI_Recv(&t1, 1, MPI_DOUBLE, 1, 0, MPI_COMM_WORLD, &status );
	   t2 = t1 - t0;
#ifdef MORE
	   printf("(%d loop) T0:%f T1:%f  Difference: %f\n", i, t0, t1, t2 );
#endif
           if ( t2 < 0 )
	   {
#ifdef DEBUG
		fprintf( stderr, "First Wtime value greater than the next.\n" );
#endif
		fprintf( stderr, "MULTI SEND/RECV REVERSED: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	   }
 	   if ( t2 == 0 )
	   {
#ifdef DEBUG
	     fprintf( stderr, "First Wtime value equal to the next.\n" );
#endif
	     fprintf( stderr, "MULTI SEND/RECV REVERSED: FAILED\n" );
  	     MPI_Abort( MPI_COMM_WORLD, -1);
	     exit(1);
	   }
	}
	else
	{
	   MPI_Recv(&t1, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status );
	   if ( (t0 = MPI_Wtime()) == 0 )
           {
#ifdef DEBUG
		fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "MULTI SEND/RECV REVERSED: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	   }
	   MPI_Send(&t0, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD );
	}
   }
  if ( rank == 0 )
  	fprintf(stderr, "MULTI SEND/RECV REVERSED: PASSED\n" );

  /*
   * Here is the multi-node test, I also call MPI_Barrier at the start to
   * make sure the master node will call the MPI_Wtime before the slave by
   * making the slave do more work. Also we take out the check for == 0
   * since the specification does not answer the behavior for this case.
   */
  for ( i = 0; i < SMALL_LOOP; i++ )
  {
	t0=0;t1=0;
        MPI_Barrier( MPI_COMM_WORLD );
	if ( rank == 1 )
	   for ( j = 0; j < NUM_LOOPS; j++ )
			temp = j+1-i;

	if ( (t0 = MPI_Wtime()) == 0 )
        {
#ifdef DEBUG
		fprintf( stderr, "MPI_WTIME is returning 0.\n" );
#endif
		fprintf( stderr, "MULTI NODE: FAILED\n" );
  		MPI_Abort( MPI_COMM_WORLD, -1);
		exit(1);
	}
        
	if ( rank == 0 ) {
	     MPI_Recv(&t1, 1, MPI_DOUBLE, 1, 0, MPI_COMM_WORLD, &status );
	     t2 = t1 - t0;
#ifdef MORE
	     printf("(%d loop) T0:%f T1:%f  Difference: %f\n", i, t0, t1, t2 );
#endif
             if ( t2 < 0 )
	       {
#ifdef DEBUG
		     fprintf( stderr, "First Wtime value greater than the next.\n" );
#endif
		     fprintf( stderr, "MULTI NODE: FAILED\n" );
  		     MPI_Abort( MPI_COMM_WORLD, -1);
		     exit(1);
	       }
	}
	else
	   MPI_Send(&t0, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD );
  }
  if ( rank == 0 )
  	fprintf( stderr, "MULTI NODE: PASSED\n" );
  MPI_Finalize();
  exit(1);
}
